from __future__ import annotations

from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
import pandas as pd
from anndata import AnnData
from beartype import beartype
from jax import lax
from numpyro.distributions import constraints
from numpyro.handlers import scale
from scipy import sparse

from .RNA_utils import sample_prior


def _linear_dt_grid(n_points: int, ratio: float) -> jnp.ndarray:
    """Return an array `t[0..n_points-1] ∈ [0,1]` where
    dt(i) grows linearly from dt_min to dt_max and dt_max/dt_min = *ratio*.
    The grid size is *n_points* (so dt array has length n_points-1).
    """
    ratio = float(max(ratio, 1.0))  # ensure ≥ 1
    N = n_points - 1  # number of intervals
    idx = jnp.arange(N, dtype=jnp.float32)
    weights = 1.0 + (ratio - 1.0) * idx / (N - 1)  # linear ramp 1 → ratio
    dt_norm = weights / weights.sum()  # normalise so Σ dt = 1
    times = jnp.concatenate([jnp.array([0.0], dtype=jnp.float32), jnp.cumsum(dt_norm)])
    return times

@beartype
def prepare_model_input(
    adata_rna: AnnData,
    tf_list: List[str],
    n_cells_col: str = "n_cells",
    prior_time_col: Optional[str] = None,
    batch_annotation: Optional[str] = None,
    prior_path_col: Optional[str] = None,
    terminal_states: Optional[List[str]] = None,
    initial_states: Optional[List[str]] = None,
    cluster_key: Optional[str] = None,
    prior_timespan: Optional[float] = 40,
    N_steps: int = 100,
    grid_type: str = "linear",
    grid_param: float = 1.0,
    custom_grid_fn: Optional[Callable[[int], jnp.ndarray]] = None,
):
    """
    Prepare RNA input data for the model by extracting spliced and unspliced counts,
    computing transcription factor indices, and extracting cell metadata including batch
    information from adata_rna.obs if provided.  In addition, if you supply
    `cluster_key`, `initial_states`, and `terminal_states`, this will automatically
    build a “Prior Paths” (1-based) from those labels.  If you still want to pass
    in a numeric column (`prior_path_col`) you can; that takes priority if given.

    Parameters
    ----------
    adata_rna : AnnData
        AnnData object containing RNA expression data with 'spliced' and 'unspliced' layers.
    tf_list : List[str]
        List of transcription factor names.
    n_cells_col : str, optional
        Column name in `adata_rna.obs` representing the number of cells per metacell.
    prior_time_col : Optional[str], optional
        Column name in `adata_rna.obs` containing prior pseudotimes; if provided, used
        to center the prior over t_c (default: None).
    batch_annotation : Optional[str], optional
        Column name in `adata_rna.obs` that contains batch information; if provided,
        cells are assigned batch indices based on this column.
    prior_path_col : Optional[str], optional
        (legacy) Column name in `adata_rna.obs` that contains a hard‐coded 1‐based path index;
        missing/NaN entries become −1 (inference).  Takes priority over the
        automatic cluster‐based labeling described below.
    terminal_states : Optional[List[str]], optional
        If provided (and `cluster_key` is also given), defines a list of cluster names
        that should be mapped to terminal paths.  They will be assigned raw_prior = 1..P
        in the order you list them.
    initial_states : Optional[List[str]], optional
        If provided (and `cluster_key` is also given), defines a list of cluster names
        that should be mapped to raw_prior = 0 (which then becomes prior_path = −2
        internally, i.e. “initial state”).
    cluster_key : Optional[str], optional
        Column in `adata_rna.obs` whose values are compared against `initial_states`
        or `terminal_states` to build a “Prior Path” assignment.  This is only used
        if `prior_path_col` is None.
    prior_timespan : float, optional
        Used if prior_time_col is None (default span).  Otherwise recalculated from data.
    N_steps : int, optional
        Number of grid steps (default: 100).
    grid_type : str, optional
        “linear” or “linear_dt” (default: “linear”).
    grid_param : float, optional
        Parameter for “linear_dt” spacing (ignored if grid_type=“linear”).
    custom_grid_fn : Optional[Callable[[int], jnp.ndarray]], optional
        If you want to supply your own function to build `times_norm`.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing:
          - `data`: JAX array of shape (cells, genes, 2) with stacked unspliced and spliced counts.
          - `M_c`: JAX array of shape (cells, 1, 1) holding the per‐cell metacell size.
          - `obs2sample`: one‐hot encoding of batch indices (shape cells × n_batch).
          - `batch_index`: JAX array of shape (cells,) with batch indices.
          - `tf_indices`: JAX array of TF gene indices.
          - `total_num_cells`: int, number of cells.
          - `n_batch`: int, number of batches.
          - `prior_time`: JAX array of prior times or None.
          - `prior_timespan`: float, span of prior_time or default.
          - `prior_path`: JAX int array of shape (cells,) with −1 or 0‐based path ids;
              prior_path = −2 means “initial state,” prior_path ≥ 0 means fixed to path id.
          - `unknown_idx`: Python list of ints, indices of cells with unknown prior (prior_path < 0).
          - `unknown_count`: int, number of unknown cells.
          - `times_norm`, `dt_norm`, `T_limits`, `hidden_units`, `num_paths`
    """

    def _to_dense_f32(layer):
        if sparse.issparse(layer):
            # cast to float32 and densify in one go
            return layer.astype(np.float32).toarray()
        else:
            # if it’s already an ndarray, just ensure dtype
            return np.asarray(layer, dtype=np.float32)

    # ========== 1) STACK RAW DATA ==========
    spliced = _to_dense_f32(adata_rna.layers["spliced"])
    unspliced = _to_dense_f32(adata_rna.layers["unspliced"])
    data = jnp.array(np.stack([unspliced, spliced], axis=-1))

    # ========== 2) METACELL SIZES ==========
    M_c = jnp.array(
        np.expand_dims(np.expand_dims(adata_rna.obs[n_cells_col].to_numpy(), -1), -1),
        dtype=jnp.float32,
    )
    total_num_cells = int(data.shape[0])

    # ========== 3) PRIOR TIME ==========
    if prior_time_col is not None:
        prior_time = jnp.array(
            adata_rna.obs[prior_time_col].to_numpy(), dtype=jnp.float32
        )
        prior_timespan = float(prior_time.max() - prior_time.min())
    else:
        prior_time = None

    # ========== 4) BATCH INDEX ==========
    if batch_annotation is not None:
        batches = adata_rna.obs[batch_annotation].astype(str)
        uniq = batches.unique()
        mapping = {b: i for i, b in enumerate(uniq)}
        batch_index = jnp.array(batches.map(mapping).to_numpy(), dtype=jnp.int32)
    else:
        batch_index = jnp.zeros(total_num_cells, dtype=jnp.int32)
    n_batch = int(jnp.unique(batch_index).shape[0])
    obs2sample = jax.nn.one_hot(batch_index, num_classes=n_batch, dtype=jnp.float32)

    # ========== 5) TF INDICES ==========
    is_tf = np.isin(adata_rna.var_names, tf_list)
    tf_indices = jnp.array(np.where(is_tf)[0], dtype=jnp.int32)
    hidden_units = int(jnp.sqrt(int(np.sum(is_tf))))

    # ========== 6) PRIOR PATH (either from prior_path_col OR from clusters) ==========
    if prior_path_col is not None:
        raw_prior = adata_rna.obs[prior_path_col].to_numpy()
    else:
        if (
            (cluster_key is None)
            or (initial_states is None)
            or (terminal_states is None)
        ):
            raise ValueError(
                "Either supply `prior_path_col`, or else supply `cluster_key`, "
                "`initial_states`, and `terminal_states` so that the function can build a prior."
            )

        # Initialize everything to NA (we will treat NA as “unknown” => prior_path = −1).
        raw_prior = pd.Series(
            data=[pd.NA] * total_num_cells, index=adata_rna.obs.index, dtype="object"
        )

        # 6a) Assign initial states → raw_prior = 0
        init_mask = adata_rna.obs[cluster_key].isin(initial_states)
        raw_prior.loc[init_mask] = 0

        # 6b) Assign terminal states → raw_prior = 1,2,… in the order given
        for idx, term_label in enumerate(terminal_states, start=1):
            mask = adata_rna.obs[cluster_key] == term_label
            raw_prior.loc[mask] = idx
        raw_prior = raw_prior.to_numpy()

    valid_paths = [int(v) for v in raw_prior if pd.notna(v) and int(v) > 0]
    num_paths = max(valid_paths) if valid_paths else 1

    prior_path_np = np.full((total_num_cells,), -1, dtype=np.int32)  # −1 ⇒ unknown
    for i, v in enumerate(raw_prior):
        if pd.isna(v):
            # leave as −1
            continue
        v_int = int(v)
        if v_int == 0:
            # “initial state” → prior_path = −2
            prior_path_np[i] = -2
        else:
            # 1…P → fixed to path v−1
            prior_path_np[i] = v_int - 1

    prior_path = jnp.array(prior_path_np, dtype=jnp.int32)
    unknown_idx = [int(i) for i, p in enumerate(prior_path_np) if p < 0]
    unknown_count = len(unknown_idx)

    # ========== 7) BUILD TIMEGRID ==========
    if prior_time_col is None:
        T_min = -prior_timespan / 2.0
        T_max = +prior_timespan / 2.0
    else:
        T_min = prior_time.min() - prior_timespan / 6.0
        T_max = prior_time.max() + prior_timespan / 6.0

    if grid_type == "linear":
        times_norm = jnp.linspace(0.0, 1.0, N_steps + 1)
    elif grid_type in {"linear_dt", "lin_dt"}:
        times_norm = _linear_dt_grid(N_steps + 1, ratio=float(grid_param))
    else:
        raise ValueError(f"Unknown grid_type '{grid_type}'.")

    # ——— 8.0 observed terminal‐state proportions ———
    # only cells with prior_path >= 0 are true terminal states
    known_mask = prior_path_np >= 0
    # count how many cells fell into each of the P paths:
    obs_path_counts = np.bincount(
        prior_path_np[known_mask], minlength=num_paths
    ).astype(np.int32)         # shape (num_paths,)
    n_known = int(obs_path_counts.sum())

    # convert to float proportions if you like:
    obs_path_props = obs_path_counts.astype(np.float32) / float(n_known)

    return OrderedDict(
        [
            ("data", data),
            ("M_c", M_c),
            ("obs2sample", obs2sample),
            ("batch_index", batch_index),
            ("tf_indices", tf_indices),
            ("total_num_cells", total_num_cells),
            ("n_batch", n_batch),
            ("prior_time", prior_time),
            ("prior_timespan", prior_timespan),
            ("prior_path", prior_path),
            ("unknown_idx", unknown_idx),
            ("unknown_count", unknown_count),
            ("times_norm", times_norm),
            ("T_limits", (T_min, T_max)),
            ("hidden_units", hidden_units),
            ("num_paths", num_paths),
            ("obs_path_props", obs_path_props),
        ],
    )

def soft_clip(y, lower, upper, sharpness=1.0):
    y1 = lower + jax.nn.softplus(sharpness * (y - lower)) / sharpness
    y2 = upper - jax.nn.softplus(sharpness * (upper - y1)) / sharpness
    return y2

@jax.jit
def interpolate_solution(
    sol_grid: jnp.ndarray,
    times_grid: jnp.ndarray,
    T_c: jnp.ndarray,
) -> jnp.ndarray:
    """
    Linearly interpolate a precomputed solution grid at specified cell times.

    Parameters:
    - sol_grid (jnp.ndarray): Solution values on a regular time grid, shape (L, num_paths, num_genes, 3).
    - times_grid (jnp.ndarray): 1D array of time points corresponding to sol_grid, length L.
    - T_c (jnp.ndarray): Array of arbitrary times at which to interpolate, shape (num_cells,).

    Returns:
    - jnp.ndarray: Interpolated solution at each time in T_c, shape (num_cells, num_paths, num_genes, 3).
    """
    idx0 = jnp.searchsorted(times_grid, T_c, side="right") - 1
    idx0 = jnp.clip(idx0, 0, sol_grid.shape[0] - 2)
    t0 = times_grid[idx0]
    t1 = times_grid[idx0 + 1]
    sol0 = sol_grid[idx0]
    sol1 = sol_grid[idx0 + 1]
    w = ((T_c - t0) / (t1 - t0))[..., None, None, None]
    return sol0 * (1 - w) + sol1 * w

@jax.jit
def hill_regulation(p_tf: jnp.ndarray, params: dict[str, jnp.ndarray]) -> jnp.ndarray:
    """
    Interpretable 2-layer MLP with positive outputs:
      - Hidden modules use ReLU thresholding
      - Linear readout gives gene logits, then softplus ensures positivity

    Args:
        params: dict with 'W1','b1','W2','b2'.
        x:      Array (batch_size, num_tfs) of TF levels.

    Returns:
        Array (batch_size, num_genes) of non-negative transcription rates.
    """

    # Module activations (logical combination of TFs)
    h = jax.nn.elu(jnp.dot(p_tf, params["W1"]) + params["b1"])
    # Linear read-out to genes
    logits = jnp.dot(h, params["W2"]) + params["b2"]
    # Enforce non-negativity: softplus activation
    out = jax.nn.softplus(logits)
    return out

# Define the ODE function using the MLP
@beartype
def dstate_dt(t, state, args):
    """
    Compute the derivative of the state vector for the coupled system using a neural network.

    Args:
        t: Time scalar.
        state: State vector [u_1, ..., u_G, s_1, ..., s_G].
        args: Tuple containing parameters (G, beta_g, gamma_g, nn_params, T_ON).

    Returns:
        Derivative of the state vector
    """

    alpha_0, beta_g, gamma_g, lamda, kappa, hill_params, tf_indices = args

    u = state[..., 0]  # Unspliced counts for all genes.
    s = state[..., 1]  # Spliced counts for all genes.
    p = state[..., 2]  # Proteins for all genes.

    # Clip state values to increase numerical stability:
    u = jnp.clip(u, 0, 1e3)
    s = jnp.clip(s, 0, 1e3)
    p = jnp.clip(p, 0, 1e3)

    # Compute transcription rates using the neural network
    p_tf = p[:, tf_indices]   # (batch, num_tfs)

    reg = hill_regulation(
        p_tf,
        hill_params
    )

    alpha = jnp.clip(alpha_0 * reg, 0.0, 1e3)

    du_dt = alpha - beta_g * u
    ds_dt = beta_g * u - gamma_g * s
    # Compute dp_dt only for TF indices
    dp_dt = jnp.zeros_like(u)
    dp_dt = dp_dt.at[:, tf_indices].set(
        lamda * s[:, tf_indices] - kappa * p[:, tf_indices]
    )

    dstate_dt = jnp.stack([du_dt, ds_dt, dp_dt], axis=-1)  # Shape: (3G,)
    return dstate_dt

def euler_integrator(
    times: jnp.ndarray,      # (L+1,)
    y0: jnp.ndarray,         # state at t0, shape (G,3)
    drift_fn: Any,           # f(t, y, args) -> dy/dt
    drift_args: Any,
    rng_key: jax.random.PRNGKey,
    sigma_noise: float = 1e-4,   # ← small noise scale
) -> jnp.ndarray:
    """
    Euler–Maruyama integrator: y_{n+1} = y_n + f(t_n,y_n)*dt + sigma_noise*sqrt(dt)*eps_n
    """
    dt = jnp.diff(times)  # (L,)
    # draw one Gaussian noise array per step
    eps = jax.random.normal(rng_key, shape=(dt.shape[0],) + y0.shape)  # (L, G, 3)

    def step(carry, inp):
        y_prev, t_prev = carry
        dt_i, eps_i = inp
        f      = drift_fn(t_prev, y_prev, drift_args)
        # scale the noise by sqrt(dt_i)
        noise  = sigma_noise * jnp.sqrt(dt_i)[None, None] * eps_i
        y_new  = y_prev + f * dt_i + noise
        # optional clipping for stability
        y_new  = jnp.clip(y_new, 0.0, 1e3)
        return (y_new, t_prev + dt_i), y_new

    (final, _), ys = lax.scan(step, (y0, times[0]), (dt, eps))
    return jnp.concatenate([y0[None], ys], axis=0)

# Define the complete NumPyro model
def model(
    data: Any,
    M_c: Any,
    obs2sample: Any,
    batch_index: Any,
    tf_indices: Any,
    total_num_cells: int,
    n_batch: int,
    prior_time: Any,
    prior_timespan: Any,
    prior_path: Any,
    unknown_idx: Any,
    unknown_count: int,
    times_norm: Any,
    T_limits: Any,
    hidden_units: Any,
    num_paths: int,
    obs_path_props: Any,
    return_alpha: bool = False,
    Tmax_alpha: float = 50.0,
    Tmax_beta: float = 1.0,
    splicing_rate_alpha_hyp_prior_alpha: float = 20.0,
    splicing_rate_alpha_hyp_prior_mean: float = 5.0,
    splicing_rate_mean_hyp_prior_alpha: float = 10.0,
    splicing_rate_mean_hyp_prior_mean: float = 1.0,
    degradation_rate_alpha_hyp_prior_alpha: float = 20.0,
    degradation_rate_alpha_hyp_prior_mean: float = 5.0,
    degradation_rate_mean_hyp_prior_alpha: float = 10.0,
    degradation_rate_mean_hyp_prior_mean: float = 1.0,
    transcription_rate_alpha_hyp_prior_alpha: float = 20.0,
    transcription_rate_alpha_hyp_prior_mean: float = 2.0,
    transcription_rate_mean_hyp_prior_alpha: float = 10.0,
    transcription_rate_mean_hyp_prior_mean: float = 5.0,
    lambda_alpha: float = 10.0,
    lambda_mean: float = 20.0,
    kappa_alpha: float = 1.0,
    kappa_mean: float = 2.0,
    detection_mean_hyp_prior_alpha: float = 1.0,
    detection_mean_hyp_prior_beta: float = 1.0,
    detection_hyp_prior_alpha: float = 10.0,
    detection_i_prior_alpha: float = 100.0,
    detection_gi_prior_alpha: float = 200.0,
    gene_add_alpha_hyp_prior_alpha: float = 9.0,
    gene_add_alpha_hyp_prior_beta: float = 3.0,
    gene_add_mean_hyp_prior_alpha: float = 1.0,
    gene_add_mean_hyp_prior_beta: float = 100.0,
    stochastic_v_ag_hyp_prior_alpha: float = 12.0,
    stochastic_v_ag_hyp_prior_beta: float = 3.0,
    **kwargs,
):
    """
    NumPyro model for coupled transcription and splicing dynamics.

    Args:
        data: Observed data array of shape (num_cells, num_genes, num_modalities).
        M_c: Number of cells in each metacell.
        batch_index: Array indicating batch assignments for each cell.
        tf_indices: Indices of genes that are TFs.
        total_num_cells: Number of cells in the full dataset.
        n_batch: Number of batches.
        Tmax_alpha: Alpha parameter for Tmax prior.
        Tmax_beta: Beta parameter for Tmax prior.
        ... (other hyperparameters for priors)
        key: Random number generator key.

    Returns:
        None. Defines the probabilistic model for inference.
    """

    num_cells = int(data.shape[0])
    num_genes = int(data.shape[1])
    num_modalities = int(data.shape[2])
    num_tfs = tf_indices.shape[0]
    batch_size = num_cells

    # Splicing Rates for mRNA
    splicing_alpha = numpyro.sample(
        "splicing_alpha",
        dist.Gamma(
            splicing_rate_alpha_hyp_prior_alpha,
            splicing_rate_alpha_hyp_prior_alpha / splicing_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_alpha = jnp.clip(splicing_alpha, a_min=1e-2, a_max=1e3)

    splicing_mean = numpyro.sample(
        "splicing_mean",
        dist.Gamma(
            splicing_rate_mean_hyp_prior_alpha,
            splicing_rate_mean_hyp_prior_alpha / splicing_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_mean = jnp.clip(splicing_mean, a_min=1e-2, a_max=1e3)

    beta_g = numpyro.sample(
        "beta_g",
        dist.Gamma(splicing_alpha, splicing_alpha / splicing_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )

    # Degradation Rates for mRNA
    degradation_alpha = numpyro.sample(
        "degradation_alpha",
        dist.Gamma(
            degradation_rate_alpha_hyp_prior_alpha,
            degradation_rate_alpha_hyp_prior_alpha
            / degradation_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_alpha = degradation_alpha + 0.001  # Prevent zero

    degradation_mean = numpyro.sample(
        "degradation_mean",
        dist.Gamma(
            degradation_rate_mean_hyp_prior_alpha,
            degradation_rate_mean_hyp_prior_alpha
            / degradation_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_mean = jnp.clip(degradation_mean, a_min=1e-2, a_max=1e3)

    gamma_g = numpyro.sample(
        "gamma_g",
        dist.Gamma(
            degradation_alpha, degradation_alpha / degradation_mean, validate_args=True
        )
        .expand([num_genes])
        .to_event(1),
    )

    # Translation rate for proteins
    lamda = numpyro.sample(
        "lambda",
        dist.Gamma(lambda_alpha, lambda_alpha / lambda_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )
    # Degradation rate for proteins
    kappa = numpyro.sample(
        "kappa",
        dist.Gamma(kappa_alpha, kappa_alpha / kappa_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )

    # Time Parameters
    # sample a *single* global scaling factor
    T_min, T_max = T_limits
    T_scaling = numpyro.sample("T_scaling", dist.Beta(1.0, 3.0))
    # total_span = (T_max - T_min) * T_scaling

    # # build *static‐shape* grid (always N_steps long)
    # abs_grid   = T_min + times_norm * (T_max - T_min)
    # times_grid = T_scaling * abs_grid

    total_span = (T_max - T_min) * T_scaling

    # 2) build *static‐shape* grid (always N_steps long)
    times_grid = T_min + times_norm * total_span
    # dt = dt_norm * total_span

    # convert your absolute prior_time (if any) into a [0,1] loc/scale
    if prior_time is None:
        prior_time = T_min + (T_max - T_min) / 2
        prior_time_sd = prior_timespan / 2.0
    else:
        prior_time_sd = prior_timespan / 12.0

    with numpyro.plate("cells", batch_size):
        t_c = numpyro.sample(
            "t_c",
            dist.TruncatedNormal(
                loc=prior_time,
                scale=prior_time_sd,
                low=T_limits[0],
                high=T_limits[1],
            ),
        )
        T_c = numpyro.deterministic("T_c", T_scaling * t_c)

    # ============= Expression model =============== #

    # scale of alpha
    alpha_0 = numpyro.sample(
        "alpha_0",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([num_genes]).to_event(1),
    )

    # Initial Conditions for ODE, sampling only for spliced and unspliced
    initial_state_2d = numpyro.sample(
        "initial_state_2d",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([1, num_genes, 2]).to_event(1),
    )

    # Create the third dimension containing proteins, only for TFs:
    third_dimension = jnp.zeros((1, num_genes))
    third_dimension = third_dimension.at[0, tf_indices].set(
        numpyro.sample(
            "initial_state_tf",
            dist.Gamma(1.0, 1.0, validate_args=True).expand([len(tf_indices)]),
        )
    )

    # Concatenate the dimensions to form the final initial state
    initial_state = jnp.concatenate(
        [initial_state_2d, third_dimension[..., None]], axis=-1
    )

    # 2) Tile across paths → shape = (num_paths, num_genes, 3)
    init_state_paths = numpyro.deterministic(
        "init_state_paths",
        jnp.tile(initial_state, (num_paths, 1, 1))
    )

    # ______ Initial conditions ______

    # # 1) Shared “root” init via Gamma (as before)
    # base_init = numpyro.sample(
    #     "base_init",
    #     dist.Gamma(1.0, 1.0)
    #         .expand([num_genes, 3])
    #         .to_event(2)   # shape (G,3): [u,s,p]
    # )

    # # 2) Additive noise per path, but keep it all in one shot
    # σ_noise = 0.000000001  # tune this to control spread of your inits
    # init_noise = numpyro.sample(
    #     "init_noise",
    #     dist.Normal(0.0, σ_noise)
    #         .expand([num_paths, num_genes, 3])
    #         .to_event(3)   # event dims = (P,G,3)
    # )

    # # 3) Combine and clip at zero
    # init_state_paths = numpyro.deterministic("init_state_paths",
    #     jnp.clip(
    #     base_init[None, ...] + init_noise,
    #     a_min=0.0,
    #     a_max=jnp.inf
    # ))

    # ───────── Path weights ─────────

    known_mask = prior_path >= 0  # fixed-path cells
    initial_mask = prior_path == -2  # “initial-state” cells

    one_hot_known = jax.nn.one_hot(jnp.clip(prior_path, 0, num_paths - 1), num_paths)

    # assume:
    #   batch_size = num_cells
    #   num_paths
    #   obs_path_props  # shape (num_paths,)
    #   initial_mask   # bool array shape (batch_size,)
    #   known_mask     # bool array shape (batch_size,)

    # ——— 1) Build the per-path log-prob “means” ———
    logits_obs = jnp.log(obs_path_props + 1e-8)          # shape: (num_paths,)

    # broadcast into a full (batch_size, num_paths) matrix:
    loc_matrix = jnp.where(
        initial_mask[:, None],                         # (batch_size, 1)
        logits_obs[None, :],                            # (1, num_paths) → (batch_size, num_paths)
        0.0,                                            # scalar → (batch_size, num_paths)
    )

    # ——— 2) Build the per-path stddevs ———
    sigma_init = 1.0
    sigma_unknown = 1.0

    # turn those scalars into vectors of length num_paths
    sigma_init_vec   = jnp.full((num_paths,), sigma_init)
    sigma_unk_vec    = jnp.full((num_paths,), sigma_unknown)

    # now broadcast into (batch_size, num_paths)
    sigma_matrix = jnp.where(
        initial_mask[:, None],                         # (batch_size, 1)
        sigma_init_vec[None, :],                       # (1, num_paths) → (batch_size, num_paths)
        sigma_unk_vec[None, :],                        # same
    )

    # ——— 3) Sample the raw logits, then softmax ———
    with numpyro.plate("cells", batch_size):
        z_pw = numpyro.sample(
          "z_pw",
          dist.Normal(loc=loc_matrix, scale=sigma_matrix)
              .to_event(1)   # event-shape = (num_paths,)
        )

    w_full = jax.nn.softmax(z_pw, axis=-1)              # (batch_size, num_paths)

    # fix the known-path cells:
    one_hot_known = jax.nn.one_hot(
      jnp.clip(prior_path, 0, num_paths-1), num_paths
    )
    path_weights = numpyro.deterministic(
      "path_weights",
      jnp.where(known_mask[:, None], one_hot_known, w_full)
    )


    avg_w = jnp.mean(path_weights, axis=0)            # shape (num_paths,)

    # 2) squared‐error against the empirical obs_path_props
    pen = jnp.sum((avg_w - obs_path_props) ** 2)

    # 3) scale to taste and inject as a factor
    penalty_scale = 1e3
    numpyro.factor("path_weight_consistency", - penalty_scale * pen)

    # Compute your scales once
    std1 = jnp.sqrt(2.0 / (num_tfs + hidden_units))
    std2 = jnp.sqrt(2.0 / (hidden_units + num_genes))

    # Grab two RNG subkeys
    key = numpyro.prng_key()
    k1, k2 = jax.random.split(key, 2)

    # Now define path-specific params via one numpyro.param each:
    W1 = numpyro.param(
        "W1",  # name in the trace
        jax.random.normal(k1, (num_paths, num_tfs, hidden_units)) * std1
    )  # → shape (P, T, H)

    b1 = numpyro.param(
        "b1",
        jnp.zeros((num_paths, hidden_units))
    )  # → shape (P, H)

    W2 = numpyro.param(
        "W2",
        jax.random.normal(k2, (num_paths, hidden_units, num_genes)) * std2
    )  # → shape (P, H, G)

    b2 = numpyro.param(
        "b2",
        jnp.zeros((num_paths, num_genes))
    )  # → shape (P, G)

    hill_params = {"W1": W1, "b1": b1, "W2": W2, "b2": b2}

    # Pass the whole dict (with leading path axis) into your ODE args:
    ode_args = (alpha_0, beta_g, gamma_g, lamda, kappa, hill_params, tf_indices)

    # Grab P different RNG keys
    base_key = jax.random.PRNGKey(0)
    keys = jax.random.split(base_key, num_paths)  # shape (P, 2) for two‐integers keys

    # leave out tf_indices from common_args:
    common_args = (alpha_0, beta_g, gamma_g, lamda, kappa)

    # def run_one_path(hill_params_i, rng_key):
    #     # 2) pack hill_params_i *before* tf_indices
    #     args_i = (*common_args, hill_params_i, tf_indices)
    #     return euler_integrator(times_grid, initial_state, dstate_dt, args_i, rng_key)

    # # Vmap it over the first axis of hill_params and over keys
    # sol_paths = jax.vmap(run_one_path, in_axes=(0, 0))(
    #     hill_params,
    #     keys,
    # )

    def run_one_path(hill_params_i, init_state_i, rng_key):
        # 1) pack your args in the correct order
        args_i = (alpha_0, beta_g, gamma_g, lamda, kappa,
                  hill_params_i, tf_indices)

        # 2) add a dummy leading batch axis so state.shape == (1, G, 3)
        y0 = init_state_i[jnp.newaxis, ...]

        # 3) run your Euler integrator
        sol = euler_integrator(times_grid, y0, dstate_dt, args_i, rng_key)
        # sol.shape == (L+1, 1, G, 3)

        # 4) remove the batch axis, returning (L+1, G, 3)
        return sol[:, 0, :, :]

    sol_paths = jax.vmap(run_one_path, in_axes=(0, 0, 0))(
        hill_params,       # (P, …)
        init_state_paths,  # (P, G, 3)
        keys,              # (P,)
    )  # -> (P, L+1, G, 3)

    sol_grid = jnp.transpose(sol_paths, (1, 0, 2, 3))

    sol_at_cells = interpolate_solution(sol_grid, times_grid, T_c)

    # Track as predictions_rearranged for compatibility purposes with some plotting functions
    numpyro.deterministic('predictions_rearranged', sol_at_cells)
    numpyro.deterministic('sol_at_cells', sol_at_cells)

    mu_expression = jnp.clip(jnp.sum(path_weights[:, :, None, None]
                                     * sol_at_cells[..., :2], axis=1), 10**(-5), 10**5)

    # Wiggliness penalty (sum of squared curvature)
    d1 = sol_grid[1:] - sol_grid[:-1]
    d2 = d1[1:] - d1[:-1]
    penalty = jnp.sum(jnp.square(d2))/times_norm.shape[0]
    # numpyro.factor("smoothness_penalty", - 10**3 * penalty)

    # ============= Detection efficiency of spliced and unspliced counts =============== #

    # Batch/Experiment-level scaling centered at 0.5
    with numpyro.plate("batch", n_batch):
        y_e = numpyro.sample(
            "detection_y_e",
            dist.Beta(detection_mean_hyp_prior_alpha,
                      detection_mean_hyp_prior_beta)
        )  # shape (n_batch,)

    # cell‐level scaling, centered at 1 (large variance)
    α_c = detection_hyp_prior_alpha
    with numpyro.plate("cells", batch_size):
        y_c = numpyro.sample(
            "detection_y_c",
            dist.Gamma(α_c, α_c)
        )  # shape (n_cells,)

    # Modality-level scaling, centered at 1 (moderate variance)
    detection_y_i = numpyro.sample(
        "detection_y_i",
        dist.Gamma(detection_i_prior_alpha, detection_i_prior_alpha, validate_args=True)
        .expand([1, 1, 2])
        .to_event(3),
    )

    # Modality and gene-level scaling, centered at 1 (small variance)
    detection_y_gi = numpyro.sample(
        "detection_y_gi",
        dist.Gamma(
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            validate_args=True,
        )
        .expand([1, num_genes, 2])
        .to_event(3),
    )

    # ======= Gene-specific additive component (Ambient RNA/ "Soup") for spliced and unspliced counts ====== #

    s_g_gene_add_alpha_hyp = numpyro.sample(
        "s_g_gene_add_alpha_hyp",
        dist.Gamma(
            gene_add_alpha_hyp_prior_alpha,
            gene_add_alpha_hyp_prior_beta,
            validate_args=True,
        ),
        sample_shape=(2,),
    )

    s_g_gene_add_mean = numpyro.sample(
        "s_g_gene_add_mean",
        dist.Gamma(
            gene_add_mean_hyp_prior_alpha,
            gene_add_mean_hyp_prior_beta,
            validate_args=True,
        )
        .expand([n_batch, 1, 2])
        .to_event(3),
    )

    s_g_gene_add_alpha_e_inv = numpyro.sample(
        "s_g_gene_add_alpha_e_inv",
        dist.Exponential(s_g_gene_add_alpha_hyp).expand([n_batch, 1, 2]).to_event(3),
    )

    s_g_gene_add_alpha_e = jnp.ones((1, 1)) / s_g_gene_add_alpha_e_inv**2

    s_g_gene_add = numpyro.sample(
        "s_g_gene_add",
        dist.Gamma(
            s_g_gene_add_alpha_e,
            s_g_gene_add_alpha_e / s_g_gene_add_mean,
            validate_args=True,
        )
        .expand([n_batch, num_genes, 2])
        .to_event(3),
    )

    # ========= Gene-specific overdispersion of spliced and unspliced counts ============== #

    stochastic_v_ag_hyp = numpyro.sample(
        "stochastic_v_ag_hyp",
        dist.Gamma(
            stochastic_v_ag_hyp_prior_alpha,
            stochastic_v_ag_hyp_prior_beta,
            validate_args=True,
        )
        .expand([1, 2])
        .to_event(2),
    )

    stochastic_v_ag_inv = numpyro.sample(
        "stochastic_v_ag_inv",
        dist.Exponential(stochastic_v_ag_hyp).expand([1, num_genes, 2]).to_event(3),
    )

    stochastic_v_ag = jnp.ones((1, 1)) / stochastic_v_ag_inv**2

    # ===================== Expected expression ======================= #

    additive_term = numpyro.deterministic(
        "additive_term", jnp.einsum("cb,bgi->cgi", obs2sample, s_g_gene_add)
    )

    normalizing_term = numpyro.deterministic(
        "normalizing_term",
        y_e[batch_index][:, None, None]  # batch‐efficiency per cell
        * y_c[:, None, None]              # cell‐scale
        * detection_y_i                   # modality‐level
        * detection_y_gi                  # gene‐modality‐level
        * M_c                             # metacell size
    )

    mu = numpyro.deterministic("mu", (mu_expression + additive_term) * normalizing_term)

    # ===================== DATA likelihood ======================= #

    concentration = stochastic_v_ag * M_c
    rate = concentration / mu

    # Likelihood (sampling distribution)
    # We apply a scale factor so that the negative log-likelihood is multiplied
    # by (total_num_cells / batch_size). That ensures an unbiased gradient estimate.
    scale_factor = total_num_cells / batch_size
    with scale(scale=scale_factor):
        data_target = numpyro.sample(
            "data_target", dist.GammaPoisson(concentration, rate), obs=data
        )



import numpy as np

from stormi.preprocessing import rhg_to_rh_indexing

# Test data
region_tf_gene_triplets = [
    [1, 3000, 1803],
    [1, 3027, 1803],
    [1, 3127, 1803],
    [10, 3000, 55],
    [63, 3000, 2500],
    [22239, 3298, 235],
    # [22239,  3198,   235], # Test ValueError. Entry doesn't map to region_tf_pairs
    [43689, 3298, 3085],
    [43689, 3298, 1660],
    [49982, 3299, 1000],
    [49999, 3299, 2000],
]

region_tf_pairs = [
    [1, 3000],
    [10, 3000],
    [63, 3000],
    [1, 3027],
    [1, 3127],
    [22239, 3298],
    [32434, 3298],
    [43689, 3298],
    [49982, 3299],
    [49998, 3299],
    [49999, 3299],
]

# Apply function to retrieve indices for enries of region_tf_gene_triplets in region_tf_pairs
region_tf_indices = rhg_to_rh_indexing(region_tf_gene_triplets, region_tf_pairs)

# Indexing region_tf_pairs with region_tf_indices should retrieve [R, H] pairs in the same order as region_tf_gene_triplets
region_tf_gene_triplets = np.array(region_tf_gene_triplets)
region_tf_pairs = np.array(region_tf_pairs)
#
assert (region_tf_pairs[region_tf_indices, :] == region_tf_gene_triplets[:, :2]).all()



"""Tests for file utility functions in the preprocessing subpackage."""

import os
import shutil
from pathlib import Path

import pytest

from stormi.preprocessing._file_utils import (
    check_command_availability,
    create_dir_if_not_exists,
    download_file,
    resolve_genome_urls,
)


def test_create_dir_if_not_exists(tmp_path):
    """Test that a directory is created if it doesn't exist."""
    test_dir = tmp_path / "test_dir"
    assert not test_dir.exists()

    # Call the function
    create_dir_if_not_exists(test_dir)

    # Verify the directory was created
    assert test_dir.exists()
    assert test_dir.is_dir()

    # Call again to test idempotency
    create_dir_if_not_exists(test_dir)
    assert test_dir.exists()


def test_check_command_availability():
    """Test that the command availability check works correctly."""
    # Test a command that should definitely be available
    assert check_command_availability("python") is True

    # Test a command that definitely shouldn't exist
    assert check_command_availability("this_command_does_not_exist_12345") is False


def test_resolve_genome_urls():
    """Test resolving genome URLs for different species and assemblies."""
    # Test mouse mm10 defaults
    gtf_url, chrom_sizes_url, fasta_url = resolve_genome_urls(
        "mouse", "mm10", None, None, None
    )
    assert "gencode" in gtf_url.lower()
    assert "mouse" in gtf_url.lower()
    assert "mm10" in chrom_sizes_url.lower()
    assert "mm10" in fasta_url.lower()

    # Test human hg38 defaults
    gtf_url, chrom_sizes_url, fasta_url = resolve_genome_urls(
        "human", "hg38", None, None, None
    )
    assert "gencode" in gtf_url.lower()
    assert "human" in gtf_url.lower()
    assert "hg38" in chrom_sizes_url.lower()
    assert "hg38" in fasta_url.lower()

    # Test user-provided URLs take precedence
    custom_gtf = "https://example.com/custom.gtf.gz"
    custom_chrom = "https://example.com/custom.chrom.sizes"
    custom_fasta = "https://example.com/custom.fa.gz"

    gtf_url, chrom_sizes_url, fasta_url = resolve_genome_urls(
        "mouse", "mm10", custom_gtf, custom_chrom, custom_fasta
    )
    assert gtf_url == custom_gtf
    assert chrom_sizes_url == custom_chrom
    assert fasta_url == custom_fasta

    # Test error for unknown assembly/species
    with pytest.raises(ValueError):
        resolve_genome_urls("unknown", "unknown", None, None, None)


def test_download_file(tmp_path, monkeypatch):
    """Test the download_file function with a mock subprocess call."""

    # Mock the subprocess.run function
    def mock_run(*args, **kwargs):
        class MockResult:
            returncode = 0

        # Create an empty file to simulate download
        with open(args[0].split(" ")[3], "w") as f:
            f.write("mock content")
        return MockResult()

    monkeypatch.setattr("subprocess.run", mock_run)

    # Test downloading a file
    out_path = tmp_path / "test_file.txt"
    download_file("https://example.com/test.txt", out_path)

    # Verify the file exists
    assert out_path.exists()
    assert out_path.read_text() == "mock content"

    # Test that download is skipped if file exists
    with monkeypatch.context() as m:
        # Set up a mock that would fail if called
        def failing_mock(*args, **kwargs):
            raise RuntimeError("This should not be called")

        m.setattr("subprocess.run", failing_mock)

        # Call again - should skip download
        download_file("https://example.com/test.txt", out_path)
        # If we get here, the download was skipped as expected



import shutil

import matplotlib as mpl
import matplotlib.pyplot as plt
from beartype import beartype
from beartype.typing import Any, Dict

__all__ = [
    "configure_matplotlib_style",
]


@beartype
def configure_matplotlib_style(
    override_defaults: Dict[str, Any] = {},
):
    """
    Configure the global matplotlib style.

    Accepts a dictionary to override default rcParams set in the
    `common.mplstyle` file.

    For example, to override the default figure size, use:

    ```
        {
            # Example: Set a global figure size
            "figure.figsize": (8, 6),
            # Add more global settings as needed
        },
    ```

    Args:
        override_defaults (Dict[str, Any], optional):
            Dictionary to update default rcParams. Defaults to {}.
    """
    plt.style.use("stormi.styles.common")

    if not shutil.which("latex"):
        mpl.rcParams.update(
            {
                "text.usetex": False,
            }
        )

    mpl.rcParams.update(
        override_defaults,
    )



"""Tests for the _motifs module."""

from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest

from stormi.preprocessing._motifs import (
    compute_in_silico_chipseq,
    compute_motif_scores,
)


@pytest.fixture
def mock_atac_matrix():
    """Create a mock ATAC matrix (cells x peaks)."""
    n_cells, n_peaks = 10, 5
    # Simulate some counts, ensure variability
    matrix = np.random.poisson(1, size=(n_cells, n_peaks)).astype(float)
    matrix[matrix > 5] = 5  # Cap counts
    # Ensure some zeros
    zero_indices = np.random.choice(
        n_cells * n_peaks, size=int(0.2 * n_cells * n_peaks), replace=False
    )
    matrix.flat[zero_indices] = 0
    # Add some noise
    matrix += np.random.normal(0, 0.1, size=matrix.shape)
    matrix = np.clip(matrix, 0, None)  # Ensure non-negative
    # Ensure some std dev for z-scoring
    col_std = matrix.std(axis=0)
    for j in range(n_peaks):
        if col_std[j] < 1e-6:
            matrix[:, j] += np.random.normal(0, 0.5, size=n_cells)
    matrix = np.clip(matrix, 0, None)  # Ensure non-negative again
    return matrix


@pytest.fixture
def mock_rna_matrix(mock_df_motif):  # Depends on mock_df_motif for TF names
    """Create a mock RNA matrix (cells x TFs)."""
    n_cells = 10  # Match mock_atac_matrix
    tf_names = mock_df_motif.columns.tolist()
    n_tfs = len(tf_names)
    # Simulate expression, ensure variability
    matrix = np.random.gamma(2, 1, size=(n_cells, n_tfs)).astype(float)
    # Ensure some zeros
    zero_indices = np.random.choice(
        n_cells * n_tfs, size=int(0.1 * n_cells * n_tfs), replace=False
    )
    matrix.flat[zero_indices] = 0
    # Ensure some std dev for z-scoring
    col_std = matrix.std(axis=0)
    for j in range(n_tfs):
        if col_std[j] < 1e-6:
            matrix[:, j] += np.random.normal(0, 0.5, size=n_cells)
    matrix = np.clip(matrix, 0, None)  # Ensure non-negative

    return matrix


@pytest.fixture
def mock_df_motif():
    """Create mock motif scores DataFrame (peaks x TFs)."""
    peaks = [f"peak{i}" for i in range(5)]
    tfs = ["TF1", "TF2", "TF3"]
    # Simulate scores after MinMax scaling (0 to 1)
    scores = np.random.rand(len(peaks), len(tfs))
    return pd.DataFrame(scores, index=peaks, columns=tfs)


@pytest.fixture
def mock_pwms_sub():
    """Create a mock pwms_sub dictionary."""
    # This would typically contain PWM objects, but for mocking we can use placeholders
    return {
        "MOTIF TF1": MagicMock(),
        "MOTIF TF2": MagicMock(),
        "MOTIF TF3": MagicMock(),
    }


@pytest.fixture
def mock_key_to_tf():
    """Create a mock key_to_tf dictionary."""
    return {
        "MOTIF TF1": "TF1",
        "MOTIF TF2": "TF2",
        "MOTIF TF3": "TF3",
    }


def test_compute_motif_scores(tmp_path, mock_pwms_sub, mock_key_to_tf):
    """Test compute_motif_scores function with minimal mocking.

    Creates a real BED file and mocks only the tangermeme functions.
    """
    # Set up test parameters
    n_peaks = 5
    n_tfs = 3
    peak_names = [f"peak{i}" for i in range(n_peaks)]
    tf_names = [f"TF{i + 1}" for i in range(n_tfs)]

    # Create a real BED file
    bed_file = tmp_path / "test.bed"
    with open(bed_file, "w") as f:
        for i, name in enumerate(peak_names):
            f.write(f"chr1\t{i * 1000}\t{i * 1000 + 200}\t{name}\n")

    # Create an empty mock FASTA file
    fasta_file = tmp_path / "test.fa"
    fasta_file.touch()

    # Create patch context managers
    extract_loci_patch = patch("tangermeme.io.extract_loci")
    fimo_patch = patch("tangermeme.tools.fimo.fimo")

    # Start patches
    with extract_loci_patch as mock_extract_loci, fimo_patch as mock_fimo:
        # Mock extract_loci: return a mock with float() method
        mock_sequence_data = MagicMock()
        mock_sequence_data.float.return_value = np.random.rand(n_peaks, 100)
        mock_extract_loci.return_value = mock_sequence_data

        # Mock fimo: return a list of DataFrames with hits
        mock_fimo_hits = []
        for motif_key, tf_name in mock_key_to_tf.items():
            # Create hits for about half the peaks for each TF
            hit_indices = np.random.choice(range(n_peaks), n_peaks // 2, replace=False)
            hit_data = {
                "motif_name": [motif_key] * len(hit_indices),
                "sequence_name": hit_indices,
                "score": np.random.rand(len(hit_indices)) * 10,
            }
            mock_fimo_hits.append(pd.DataFrame(hit_data))
        mock_fimo.return_value = mock_fimo_hits

        # Call the function
        result_df = compute_motif_scores(
            bed_file=bed_file,
            fasta_file=fasta_file,
            pwms_sub=mock_pwms_sub,
            key_to_tf=mock_key_to_tf,
            n_peaks=n_peaks,
            window=100,
            threshold=1e-4,
            batch_size_Fimo_scan=100,
        )

        # Verify result
        assert isinstance(result_df, pd.DataFrame)
        assert result_df.shape == (n_peaks, n_tfs)
        assert list(result_df.index) == peak_names
        assert set(result_df.columns) == set(tf_names)

        # Check scores are in range [0, 1] due to MinMaxScaler
        assert result_df.values.min() >= 0.0
        assert result_df.values.max() <= 1.0

        # Verify mocks were called appropriately
        mock_extract_loci.assert_called_once()
        mock_fimo.assert_called_once()


def test_compute_in_silico_chipseq(mock_df_motif, mock_atac_matrix, mock_rna_matrix):
    """Test compute_in_silico_chipseq function."""
    correlation_percentile = 95.0
    n_bg_peaks_for_corr = 3  # Use small value to ensure it doesn't exceed n_peaks
    n_peaks, n_tfs = mock_df_motif.shape
    n_cells = mock_atac_matrix.shape[0]

    # Ensure matrix dimensions align
    assert mock_atac_matrix.shape == (n_cells, n_peaks)
    assert mock_rna_matrix.shape == (n_cells, n_tfs)
    assert list(mock_df_motif.columns) == [
        f"TF{i + 1}" for i in range(n_tfs)
    ]  # Check TF order if needed

    result_df = compute_in_silico_chipseq(
        df_motif=mock_df_motif,
        atac_matrix=mock_atac_matrix,
        rna_matrix=mock_rna_matrix,
        correlation_percentile=correlation_percentile,
        n_bg_peaks_for_corr=n_bg_peaks_for_corr,
    )

    assert isinstance(result_df, pd.DataFrame)

    # Check columns (long format expected)
    expected_columns = ["peak_name", "Motif_name", "Matching_Score"]
    assert all(col in result_df.columns for col in expected_columns)
    assert len(result_df.columns) == len(expected_columns)

    # Check data types
    assert pd.api.types.is_string_dtype(result_df["peak_name"])
    assert pd.api.types.is_string_dtype(result_df["Motif_name"])
    assert pd.api.types.is_numeric_dtype(result_df["Matching_Score"])

    # Check score range (scaled between -1 and 1)
    if not result_df.empty:
        assert (
            result_df["Matching_Score"].min() >= -1.0 - 1e-9
        )  # Allow for float precision
        assert (
            result_df["Matching_Score"].max() <= 1.0 + 1e-9
        )  # Allow for float precision

    # Check that peak names and motif names come from the input df_motif
    if not result_df.empty:
        assert set(result_df["peak_name"]).issubset(set(mock_df_motif.index))
        assert set(result_df["Motif_name"]).issubset(set(mock_df_motif.columns))

    # Check that only non-zero scores are returned
    assert not np.any(np.isclose(result_df["Matching_Score"], 0))



import anndata as ad
import numpy as np
import pandas as pd
import pytest
import scanpy as sc
from matplotlib.figure import Figure

###############################################################################
# Monkey-patch Gamma distribution to skip validate_args in tests
from numpyro.distributions.continuous import Gamma

_orig_gamma_init = Gamma.__init__


def _no_validate_gamma_init(self, concentration, rate, validate_args=True):
    _orig_gamma_init(self, concentration, rate, validate_args=False)


Gamma.__init__ = _no_validate_gamma_init
###############################################################################

from numpyro.infer.initialization import init_to_mean

import stormi.posterior as posterior_module
from stormi.guides.AmortizedNormal import AmortizedNormal, warm_up_guide
from stormi.models.RNA_2layers_MultiplePaths_SDE import (
    model,
    prepare_model_input,
)
from stormi.plotting import plot_elbo_loss, posterior_data_geneset_multipath
from stormi.train import train_svi_jit

###############################################################################
# Fixtures
###############################################################################


@pytest.fixture(scope="module")
def adata_rna():
    """
    Synthetic AnnData with a flexible number of cells × genes,
    including all obs columns needed for prepare_model_input.
    """
    np.random.seed(0)
    n_cells = 5
    n_genes = 5

    genes = [f"gene{i}" for i in range(1, n_genes + 1)]
    cells = [f"cell{i}" for i in range(1, n_cells + 1)]

    # Simulate expression counts (we'll treat both spliced/unspliced the same here)
    X = np.random.randint(0, 100, size=(n_cells, n_genes))

    # Construct obs with required columns:
    n_cells_col = np.random.randint(1, 11, size=n_cells)
    day_float = np.random.rand(n_cells) * 100

    # Ensure at least one “Mesoderm” and one “CM”
    available_batches = ["expA", "expB", "expC"]
    exper = ["expA", "expB"] if n_cells >= 2 else ["expA"]
    if len(exper) < n_cells:
        exper += list(np.random.choice(available_batches, size=n_cells - len(exper)))
    exper = exper[:n_cells]

    cluster_labels = ["Mesoderm", "CM"] if n_cells >= 2 else ["Mesoderm"]
    all_states = ["Mesoderm", "CM", "EC", "Fib"]
    if len(cluster_labels) < n_cells:
        cluster_labels += list(
            np.random.choice(all_states, size=n_cells - len(cluster_labels))
        )
    cluster_labels = cluster_labels[:n_cells]

    obs = pd.DataFrame(
        {
            "n_cells": n_cells_col,
            "day_float": day_float,
            "experiment": exper,
            "RNA_cell_type": cluster_labels,
        },
        index=cells,
    )

    var = pd.DataFrame(index=genes)
    adata = ad.AnnData(X=X, obs=obs, var=var)
    # Duplicate X into “spliced” and “unspliced” layers so that prepare_model_input sees a 3D structure
    adata.layers["spliced"] = X.copy()
    adata.layers["unspliced"] = X.copy()
    adata.raw = adata
    return adata


@pytest.fixture(scope="module")
def tf_list():
    return ["gene1", "gene2"]


@pytest.fixture(scope="module")
def model_input(adata_rna, tf_list):
    return prepare_model_input(
        adata_rna,
        tf_list,
        n_cells_col="n_cells",
        prior_time_col="day_float",
        batch_annotation="experiment",
        terminal_states=["CM", "EC", "Fib"],
        initial_states=["Mesoderm"],
        cluster_key="RNA_cell_type",
    )


@pytest.fixture(scope="module")
def guide(model_input):
    """
    Build AmortizedNormal guide via warm-up.
    """
    warm_params = warm_up_guide(model, model_input)
    return AmortizedNormal(model, init_net_params=warm_params, init_loc_fn=init_to_mean)


@pytest.fixture(scope="module")
def svi_results(guide, model_input):
    """
    Run a couple of SVI iterations and return:
    (guide, svi, svi_state, losses, updated_model_input)
    """
    num_cells = model_input["data"].shape[0]
    guide_out, svi, svi_state, losses, updated_mi = train_svi_jit(
        model,
        guide,
        model_input,
        max_iterations=2,
        cell_batch_size=4,
        log_interval=1,
        min_lr=1e-3,
        max_lr=1e-2,
        ramp_up_fraction=0.5,
    )
    return guide_out, svi, svi_state, losses, updated_mi


###############################################################################
# Tests
###############################################################################


@pytest.mark.slow
def test_svi_losses(svi_results):
    """
    Ensure losses is a non-empty list.
    """
    _, _, _, losses, _ = svi_results
    assert isinstance(losses, list)
    assert len(losses) > 0


def test_plot_elbo_loss(svi_results):
    """
    Test that plot_elbo_loss returns either None or a Figure.
    """
    _, _, _, losses, _ = svi_results
    fig = plot_elbo_loss(losses)
    assert fig is None or isinstance(fig, Figure)


def test_posterior_extraction(svi_results, adata_rna, tf_list):
    """
    Extract posterior
    """
    guide_out, svi, svi_state, _, mi = svi_results
    posterior = posterior_module.extract_posterior_estimates(
        model,
        guide_out,
        svi,
        svi_state,
        quantiles=[],
        num_samples=0,
        modes=[1],
        model_input=mi,
        deterministic_sites=[
            "T_c",
            "mu",
            "normalizing_term",
            "additive_term",
            "sol_at_cells",
            "path_weights",
        ],
    )



"""Tests for the _metacells module."""

import numpy as np
import pandas as pd
import pytest
import scipy.sparse as sp
from anndata import AnnData

from stormi.preprocessing import compute_metacells, convert_to_dense


@pytest.fixture
def simple_adata_rna_with_clusters():
    """Create a small AnnData object with clusters for testing."""
    n_cells, n_genes = 50, 100
    X = np.random.rand(n_cells, n_genes)
    adata = AnnData(X=X)

    # Add clustering information
    adata.obs["leiden"] = np.repeat(["0", "1", "2", "3", "4"], 10)

    # Add latent representation
    adata.obsm["X_scVI"] = np.random.rand(n_cells, 10)

    return adata


@pytest.fixture
def simple_adata_atac_with_clusters():
    """Create a small ATAC AnnData object with the same cells and clusters."""
    n_cells, n_peaks = 50, 200
    X = np.random.rand(n_cells, n_peaks)
    adata = AnnData(X=X)

    # Add clustering information
    adata.obs["leiden"] = np.repeat(["0", "1", "2", "3", "4"], 10)

    return adata


def test_convert_to_dense():
    """Test converting sparse matrices to dense arrays."""
    # Test with dense array
    dense_array = np.random.rand(10, 20)
    result_dense = convert_to_dense(dense_array)
    assert np.array_equal(result_dense, dense_array)

    # Test with sparse matrix
    sparse_matrix = sp.csr_matrix(dense_array)
    result_sparse = convert_to_dense(sparse_matrix)
    assert np.array_equal(result_sparse, dense_array)


@pytest.mark.skipif(
    True,
    reason="Skipping metacell tests as they require scanpy/clustering dependencies",
)
def test_compute_metacells_rna_only(simple_adata_rna_with_clusters):
    """Test computing metacells from RNA data only."""
    # Make the test small for speed
    simple_adata_rna_with_clusters = simple_adata_rna_with_clusters[:20].copy()

    # Run compute_metacells
    metacells = compute_metacells(
        adata_rna=simple_adata_rna_with_clusters,
        latent_key="X_scVI",
    )

    # Check metacell structure
    assert isinstance(metacells, AnnData)
    assert metacells.n_obs <= simple_adata_rna_with_clusters.n_obs
    assert metacells.n_vars == simple_adata_rna_with_clusters.n_vars
    assert "n_cells" in metacells.obs


@pytest.mark.skipif(
    True,
    reason="Skipping metacell tests as they require scanpy/clustering dependencies",
)
def test_compute_metacells_multimodal(
    simple_adata_rna_with_clusters, simple_adata_atac_with_clusters
):
    """Test computing metacells from both RNA and ATAC data."""
    # Make the test small for speed
    simple_adata_rna_with_clusters = simple_adata_rna_with_clusters[:20].copy()
    simple_adata_atac_with_clusters = simple_adata_atac_with_clusters[:20].copy()

    # Use the same observation names for RNA and ATAC
    obs_names = [f"cell_{i}" for i in range(20)]
    simple_adata_rna_with_clusters.obs_names = obs_names
    simple_adata_atac_with_clusters.obs_names = obs_names

    # Run compute_metacells
    metacells_rna, metacells_atac = compute_metacells(
        adata_rna=simple_adata_rna_with_clusters,
        adata_atac=simple_adata_atac_with_clusters,
        latent_key="X_scVI",
    )

    # Check metacell structure
    assert isinstance(metacells_rna, AnnData)
    assert isinstance(metacells_atac, AnnData)
    assert metacells_rna.n_obs == metacells_atac.n_obs
    assert metacells_rna.n_vars == simple_adata_rna_with_clusters.n_vars
    assert metacells_atac.n_vars == simple_adata_atac_with_clusters.n_vars
    assert "n_cells" in metacells_rna.obs
    assert "n_cells" in metacells_atac.obs



"""Functions for computing data complexity and auto-selecting model parameters."""

import math

import numpy as np
import scanpy as sc
from anndata import AnnData


def compute_pca_complexity(
    adata: AnnData, variance_threshold: float = 0.9, n_comps: int = 50
) -> int:
    """
    Computes a simple measure of dataset complexity: the number of principal components
    required to explain at least `variance_threshold` of the variance.

    If PCA has not yet been computed, this function runs sc.pp.pca on the adata.
    """
    # Ensure n_comps doesn't exceed dataset dimensions
    n_comps = min(n_comps, adata.n_obs - 1, adata.n_vars - 1)

    if "pca" not in adata.uns:
        sc.pp.pca(adata, n_comps=n_comps)
    variance_ratio = adata.uns["pca"]["variance_ratio"]
    cum_var = np.cumsum(variance_ratio)
    complexity = int(np.searchsorted(cum_var, variance_threshold)) + 1
    return complexity


def default_n_latent(n_obs: int, complexity: int) -> int:
    """Compute a default latent dimension for scVI/multiVI models.

    Base is 20 + 10 * log10(n_obs/1e3) and then adjusted upward by 0.5 * complexity.
    Capped at 150.

    Args:
        n_obs: Number of observations (cells).
        complexity: Dataset complexity measure.

    Returns:
        int: Recommended latent dimension size.
    """
    base = 20 + 10 * math.log10(n_obs / 1000)
    return int(max(20, min(150, base + 0.5 * complexity)))


def default_n_hidden(n_obs: int, complexity: int) -> int:
    """Compute a default number of hidden units per layer.

    Base is 256 + 64 * log10(n_obs/1e3) and then adjusted upward by 8 * complexity.
    Capped at 1024.

    Args:
        n_obs: Number of observations (cells).
        complexity: Dataset complexity measure.

    Returns:
        int: Recommended number of hidden units.
    """
    base = 256 + 64 * math.log10(n_obs / 1000)
    return int(max(256, min(1024, base + 8 * complexity)))


def default_n_layers(n_obs: int, complexity: int) -> int:
    """Return a default number of layers for neural network models.

    For fewer than 1e5 cells, use 2 layers if complexity < 20, else 3.
    For larger datasets, use 3 layers if complexity < 30, else 4.

    Args:
        n_obs: Number of observations (cells).
        complexity: Dataset complexity measure.

    Returns:
        int: Recommended number of layers.
    """
    if n_obs < 1e5:
        return 2 if complexity < 20 else 3
    else:
        return 3 if complexity < 30 else 4


def default_epochs(n_obs: int, complexity: int) -> int:
    """Compute a default number of training epochs.

    Base increases with n_obs and is scaled by the complexity.
    For 1e4 cells with moderate complexity, ~600 epochs are used.
    The final number is increased by a factor (1 + complexity/50)
    to ensure higher iterations for more complex datasets.

    Args:
        n_obs: Number of observations (cells).
        complexity: Dataset complexity measure.

    Returns:
        int: Recommended number of training epochs.
    """
    base = 600 + 200 * math.log10(n_obs / 10000)
    return int(max(400, base * (1 + complexity / 50)))



from __future__ import annotations

from typing import Any, Dict

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from beartype import beartype
from numpyro.handlers import scale

from .RNA_utils import prepare_model_input, sample_prior
from .utils import solve_DE, sort_times_over_all_cells


@beartype
def mlp(params: dict, x: jnp.ndarray) -> jnp.ndarray:
    """
    Multilayer Perceptron (MLP) with residual connections.

    Args:
        params: Dictionary containing neural network parameters (weights and biases).
        x: Input data array.

    Returns:
        Output array after passing through the MLP.
    """

    # Block 1
    h = x @ params["W1"] + params["b1"]  # (n_cells, 512)
    h = jax.nn.softplus(h)
    h = h @ params["W2"] + params["b2"]  # (n_cells, 1000)
    h = jax.nn.softplus(h)
    h = h + x  # residual

    # Final head
    h2 = h @ params["W3"] + params["b3"]  # (n_cells, 512)
    h2 = jax.nn.softplus(h2)
    out = h2 @ params["W4"] + params["b4"]  # (n_cells, 2000)

    return out


# Define the ODE function using the MLP
@beartype
def dstate_dt(t, state, args):
    """
    Compute the derivative of the state vector for the coupled system using a neural network.

    Args:
        t: Time scalar.
        state: State vector [u_1, ..., u_G, s_1, ..., s_G].
        args: Tuple containing parameters (G, beta_g, gamma_g, nn_params, T_ON).

    Returns:
        Derivative of the state vector
    """

    alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices, T_ON = args
    u = state[..., 0]  # Unspliced counts for all genes.
    s = state[..., 1]  # Spliced counts for all genes.
    p = state[..., 2]  # Proteins for all genes.

    # Clip state values to increase numerical stability:
    u = jnp.clip(u, 0, 1e3)
    s = jnp.clip(s, 0, 1e3)
    p = jnp.clip(p, 0, 1e3)

    # Compute transcription rates using the neural network
    alpha = alpha_0 * jnp.clip(mlp(nn_params, p[:, tf_indices]), 0, 1e3)  # Shape: (G,)

    du_dt = alpha - beta_g * u
    ds_dt = beta_g * u - gamma_g * s
    # Compute dp_dt only for TF indices
    dp_dt = jnp.zeros_like(u)
    dp_dt = dp_dt.at[:, tf_indices].set(
        lamda * s[:, tf_indices] - kappa * p[:, tf_indices]
    )

    dstate_dt = jnp.stack([du_dt, ds_dt, dp_dt], axis=-1)  # Shape: (3G,)
    return dstate_dt


# Define the complete NumPyro model
@beartype
def model(
    data: Any,
    M_c: Any,
    batch_index: Any,
    tf_indices: Any,
    total_num_cells: int,
    n_batch: int,
    prior_time: Any,
    prior_timespan: Any,
    prior_path: Any,
    unknown_idx: Any,
    num_paths: int,
    return_alpha: bool = False,
    Tmax_alpha: float = 50.0,
    Tmax_beta: float = 1.0,
    splicing_rate_alpha_hyp_prior_alpha: float = 20.0,
    splicing_rate_alpha_hyp_prior_mean: float = 5.0,
    splicing_rate_mean_hyp_prior_alpha: float = 10.0,
    splicing_rate_mean_hyp_prior_mean: float = 1.0,
    degradation_rate_alpha_hyp_prior_alpha: float = 20.0,
    degradation_rate_alpha_hyp_prior_mean: float = 5.0,
    degradation_rate_mean_hyp_prior_alpha: float = 10.0,
    degradation_rate_mean_hyp_prior_mean: float = 1.0,
    transcription_rate_alpha_hyp_prior_alpha: float = 20.0,
    transcription_rate_alpha_hyp_prior_mean: float = 2.0,
    transcription_rate_mean_hyp_prior_alpha: float = 10.0,
    transcription_rate_mean_hyp_prior_mean: float = 5.0,
    lambda_alpha: float = 1.0,
    lambda_mean: float = 1.0,
    kappa_alpha: float = 1.0,
    kappa_mean: float = 1.0,
    detection_mean_hyp_prior_alpha: float = 1.0,
    detection_mean_hyp_prior_beta: float = 1.0,
    detection_hyp_prior_alpha: float = 10.0,
    detection_i_prior_alpha: float = 100.0,
    detection_gi_prior_alpha: float = 200.0,
    gene_add_alpha_hyp_prior_alpha: float = 9.0,
    gene_add_alpha_hyp_prior_beta: float = 3.0,
    gene_add_mean_hyp_prior_alpha: float = 1.0,
    gene_add_mean_hyp_prior_beta: float = 100.0,
    stochastic_v_ag_hyp_prior_alpha: float = 9.0,
    stochastic_v_ag_hyp_prior_beta: float = 3.0,
    sde_rng_key=0,
):
    """
    NumPyro model for coupled transcription and splicing dynamics.

    Args:
        data: Observed data array of shape (num_cells, num_genes, num_modalities).
        M_c: Number of cells in each metacell.
        batch_index: Array indicating batch assignments for each cell.
        tf_indices: Indices of genes that are TFs.
        total_num_cells: Number of cells in the full dataset.
        n_batch: Number of batches.
        Tmax_alpha: Alpha parameter for Tmax prior.
        Tmax_beta: Beta parameter for Tmax prior.
        ... (other hyperparameters for priors)
        key: Random number generator key.

    Returns:
        None. Defines the probabilistic model for inference.
    """

    num_cells = int(data.shape[0])
    num_genes = int(data.shape[1])
    num_modalities = int(data.shape[2])
    num_tfs = tf_indices.shape[0]
    batch_size = num_cells
    obs2sample = jax.nn.one_hot(
        batch_index, num_classes=n_batch
    )  # Shape: (num_cells, n_batch)

    # Splicing Rates for mRNA
    splicing_alpha = numpyro.sample(
        "splicing_alpha",
        dist.Gamma(
            splicing_rate_alpha_hyp_prior_alpha,
            splicing_rate_alpha_hyp_prior_alpha / splicing_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_alpha = jnp.clip(splicing_alpha, a_min=1e-2, a_max=1e3)

    splicing_mean = numpyro.sample(
        "splicing_mean",
        dist.Gamma(
            splicing_rate_mean_hyp_prior_alpha,
            splicing_rate_mean_hyp_prior_alpha / splicing_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_mean = jnp.clip(splicing_mean, a_min=1e-2, a_max=1e3)

    beta_g = numpyro.sample(
        "beta_g",
        dist.Gamma(splicing_alpha, splicing_alpha / splicing_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )

    # Degradation Rates for mRNA
    degradation_alpha = numpyro.sample(
        "degradation_alpha",
        dist.Gamma(
            degradation_rate_alpha_hyp_prior_alpha,
            degradation_rate_alpha_hyp_prior_alpha
            / degradation_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_alpha = degradation_alpha + 0.001  # Prevent zero

    degradation_mean = numpyro.sample(
        "degradation_mean",
        dist.Gamma(
            degradation_rate_mean_hyp_prior_alpha,
            degradation_rate_mean_hyp_prior_alpha
            / degradation_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_mean = jnp.clip(degradation_mean, a_min=1e-2, a_max=1e3)

    gamma_g = numpyro.sample(
        "gamma_g",
        dist.Gamma(
            degradation_alpha, degradation_alpha / degradation_mean, validate_args=True
        )
        .expand([num_genes])
        .to_event(1),
    )

    # Translation rate for proteins
    lamda = numpyro.sample(
        "lambda",
        dist.Gamma(lambda_alpha, lambda_alpha / lambda_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )
    # Degradation rate for proteins
    kappa = numpyro.sample(
        "kappa",
        dist.Gamma(kappa_alpha, kappa_alpha / kappa_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )

    # Time Parameters
    prior_time_sd = prior_timespan / 6
    T_scaling = numpyro.sample("T_scaling", dist.Gamma(1.0, 1.0))
    if prior_time is None:
        with numpyro.plate("cells", batch_size):
            lower_limit = -prior_timespan / 2
            upper_limit = prior_timespan / 2
            t_c = numpyro.sample(
                "t_c",
                dist.TruncatedNormal(
                    low=lower_limit, high=upper_limit, loc=0, scale=prior_time_sd
                ),
            )
            T_c = numpyro.deterministic(
                "T_c", jnp.clip(t_c, min=lower_limit, max=upper_limit)
            )
    else:
        with numpyro.plate("cells", batch_size):
            lower_limit = jnp.min(prior_time) - 3 * prior_time_sd
            upper_limit = jnp.max(prior_time) + 3 * prior_time_sd
            t_c = numpyro.sample(
                "t_c",
                dist.TruncatedNormal(
                    low=-4 * prior_time_sd,
                    high=4 * prior_time_sd,
                    loc=0,
                    scale=prior_time_sd,
                ),
            )
            T_c = numpyro.deterministic("T_c", T_scaling * (t_c + prior_time))
    T_ON = lower_limit - 1

    # ============= Expression model =============== #

    # scale of alpha
    alpha_0 = numpyro.sample(
        "alpha_0",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([num_genes]).to_event(1),
    )

    # Hyper-prior for the shared “center” of (unspliced, spliced)
    init_center_2d = numpyro.sample(
        "init_center_2d",
        dist.Gamma(2.0, 0.5).expand([num_genes, 2]).to_event(2),
    )

    # Hyper-prior for the shared center of TF-protein initials
    init_center_tf = numpyro.sample(
        "init_center_tf",
        dist.Gamma(2.0, 0.5).expand([len(tf_indices)]).to_event(1),
    )

    # Small Gaussian jitter around zero for each path
    σ_dev = 1.0
    jitter_2d = numpyro.sample(
        "jitter_2d",
        dist.Normal(0.0, σ_dev).expand([num_paths, num_genes, 2]).to_event(3),
    )
    jitter_tf = numpyro.sample(
        "jitter_tf",
        dist.Normal(0.0, σ_dev).expand([num_paths, len(tf_indices)]).to_event(2),
    )

    # Build each path’s initial state by adding jitter to the shared center
    initial_state_2d = init_center_2d[None, :, :] + jitter_2d
    # → shape (num_paths, num_genes, 2)

    # Now insert the TF-protein entries into the third dim
    third_dim = jnp.zeros((num_paths, num_genes))
    third_dim = third_dim.at[:, tf_indices].set(init_center_tf[None, :] + jitter_tf)
    initial_state = jnp.concatenate(
        [initial_state_2d, third_dim[..., None]],
        axis=-1,
    )

    initial_state = numpyro.deterministic(
        "initial_state",
        jnp.concatenate(
            [initial_state_2d, third_dim[..., None]],
            axis=-1,
        ),
    )

    # Neural Net parameters
    rng = jax.random.PRNGKey(0)
    P = numpyro.param

    nn_params = {
        # Block 1
        "W1": P("W1", jax.random.normal(rng, (num_tfs, 512)) * 0.01),
        "b1": P("b1", jnp.zeros((512,))),
        "W2": P("W2", jax.random.normal(rng, (512, num_tfs)) * 0.01),
        "b2": P("b2", jnp.zeros((num_tfs,))),
        # Final head
        "W3": P("W3", jax.random.normal(rng, (num_tfs, 512)) * 0.01),
        "b3": P("b3", jnp.zeros((512,))),
        "W4": P("W4", jax.random.normal(rng, (512, num_genes)) * 0.01),
        "b4": P("b4", jnp.zeros((num_genes,))),
    }

    # Prepare Parameters for ODE Solver
    params = (alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices, T_ON)

    # Get Ordered Time Vector
    all_times, time_indices, _ = sort_times_over_all_cells(T_c)

    # Solve the Coupled ODE
    predictions = solve_DE(
        ts=all_times.squeeze(),
        params=params,
        initial_state=initial_state,
        time_step=0.1,
        drift=dstate_dt,
        t0=lower_limit,
        t1=upper_limit,
    )

    predictions = numpyro.deterministic("predictions", predictions.astype(jnp.float32))

    predictions_rearranged = numpyro.deterministic(
        "predictions_rearranged", predictions[time_indices.ravel(), :]
    )

    # —— path_weights —— #

    # one-hot for the cells with known path (works even if prior_path is None)
    if prior_path is None:
        one_hot = jnp.zeros((batch_size, num_paths))
        unknown_idx = list(range(batch_size))
    else:
        one_hot = jax.nn.one_hot(prior_path, num_paths)  # (N, K)
        unknown_idx = unknown_idx  # comes in as a Python list

    n_unknown = len(unknown_idx)

    # sample only for unknown cells
    w_unknown = numpyro.sample(
        "path_weights_unknown",
        dist.Dirichlet(jnp.ones(num_paths)).expand([n_unknown]).to_event(1),
    )  # shape (n_unknown, K)

    # scatter them back into a full array
    path_weights = one_hot
    for j, i in enumerate(unknown_idx):
        path_weights = path_weights.at[i].set(w_unknown[j])

    numpyro.deterministic("path_weights", path_weights)  # (N, K)

    # AVERAGE THE predictions REARRANGED BY THESE WEIGHTS ---
    weighted_preds = jnp.sum(
        predictions_rearranged * path_weights[:, :, None, None], axis=1
    )  # → (num_cells, num_genes, 3)

    # now downstream we use weighted_preds as mu_expression
    mu_expression = jnp.clip(
        weighted_preds[..., :2],
        a_min=1e-5,
        a_max=1e5,
    )

    # if return_alpha:
    #     final_p_clipped = jnp.clip(predictions_rearranged[..., 2].squeeze(1), 0, 1e3)
    #     alpha_cg = jnp.clip(
    #         alpha_0 * mlp(nn_params, final_p_clipped[:, tf_indices]), 0, 1e3
    #     )
    #     alpha_cg = numpyro.deterministic("alpha_cg", alpha_cg)

    # ============= Detection efficiency of spliced and unspliced counts =============== #

    detection_mean_y_e = numpyro.sample(
        "detection_mean_y_e",
        dist.Beta(
            jnp.ones((1, 1)) * detection_mean_hyp_prior_alpha,
            jnp.ones((1, 1)) * detection_mean_hyp_prior_beta,
            validate_args=True,
        )
        .expand([n_batch, 1])
        .to_event(2),
    )

    beta = ((detection_hyp_prior_alpha / (obs2sample @ detection_mean_y_e)).T)[0, ...]

    with numpyro.plate("cells", batch_size):
        detection_y_c = numpyro.sample(
            "detection_y_c", dist.Gamma(detection_hyp_prior_alpha, beta)
        )

    detection_y_i = numpyro.sample(
        "detection_y_i",
        dist.Gamma(detection_i_prior_alpha, detection_i_prior_alpha, validate_args=True)
        .expand([1, 1, 2])
        .to_event(3),
    )

    detection_y_gi = numpyro.sample(
        "detection_y_gi",
        dist.Gamma(
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            validate_args=True,
        )
        .expand([1, num_genes, 2])
        .to_event(3),
    )

    # ======= Gene-specific additive component (Ambient RNA/ "Soup") for spliced and unspliced counts ====== #

    s_g_gene_add_alpha_hyp = numpyro.sample(
        "s_g_gene_add_alpha_hyp",
        dist.Gamma(
            gene_add_alpha_hyp_prior_alpha,
            gene_add_alpha_hyp_prior_beta,
            validate_args=True,
        ),
        sample_shape=(2,),
    )

    s_g_gene_add_mean = numpyro.sample(
        "s_g_gene_add_mean",
        dist.Gamma(
            gene_add_mean_hyp_prior_alpha,
            gene_add_mean_hyp_prior_beta,
            validate_args=True,
        )
        .expand([n_batch, 1, 2])
        .to_event(3),
    )

    s_g_gene_add_alpha_e_inv = numpyro.sample(
        "s_g_gene_add_alpha_e_inv",
        dist.Exponential(s_g_gene_add_alpha_hyp).expand([n_batch, 1, 2]).to_event(3),
    )

    s_g_gene_add_alpha_e = jnp.ones((1, 1)) / s_g_gene_add_alpha_e_inv**2

    s_g_gene_add = numpyro.sample(
        "s_g_gene_add",
        dist.Gamma(
            s_g_gene_add_alpha_e,
            s_g_gene_add_alpha_e / s_g_gene_add_mean,
            validate_args=True,
        )
        .expand([n_batch, num_genes, 2])
        .to_event(3),
    )

    # ========= Gene-specific overdispersion of spliced and unspliced counts ============== #

    stochastic_v_ag_hyp = numpyro.sample(
        "stochastic_v_ag_hyp",
        dist.Gamma(
            stochastic_v_ag_hyp_prior_alpha,
            stochastic_v_ag_hyp_prior_beta,
            validate_args=True,
        )
        .expand([1, 2])
        .to_event(2),
    )

    stochastic_v_ag_inv = numpyro.sample(
        "stochastic_v_ag_inv",
        dist.Exponential(stochastic_v_ag_hyp).expand([1, num_genes, 2]).to_event(3),
    )

    stochastic_v_ag = jnp.ones((1, 1)) / stochastic_v_ag_inv**2

    # ===================== Expected expression ======================= #

    additive_term = numpyro.deterministic(
        "additive_term", jnp.einsum("cb,bgi->cgi", obs2sample, s_g_gene_add)
    )

    normalizing_term = numpyro.deterministic(
        "normalizing_term",
        detection_y_c[:, None, None] * detection_y_i * detection_y_gi * M_c,
    )

    mu = numpyro.deterministic("mu", (mu_expression + additive_term) * normalizing_term)

    # ===================== DATA likelihood ======================= #

    concentration = stochastic_v_ag * M_c
    rate = concentration / mu

    # Likelihood (sampling distribution)
    # We apply a scale factor so that the negative log-likelihood is multiplied
    # by (total_num_cells / batch_size). That ensures an unbiased gradient estimate.
    scale_factor = total_num_cells / batch_size
    with scale(scale=scale_factor):
        data_target = numpyro.sample(
            "data_target", dist.GammaPoisson(concentration, rate), obs=data
        )



import logging

import pytest

from stormi.logging import configure_logging


def test_configure_logging_returns_logger():
    """Test that configure_logging returns a Logger instance."""
    logger = configure_logging()
    assert isinstance(logger, logging.Logger)
    assert logger.name == "stormi"


def test_configure_logging_custom_name():
    """Test that configure_logging uses the provided logger name."""
    logger = configure_logging("custom_name")
    assert logger.name == "custom_name"


@pytest.mark.parametrize(
    "level_name,expected_level",
    [
        ("DEBUG", logging.DEBUG),
        ("INFO", logging.INFO),
        ("WARNING", logging.WARNING),
        ("ERROR", logging.ERROR),
        ("CRITICAL", logging.CRITICAL),
    ],
)
def test_configure_logging_respects_environment_variable(
    monkeypatch, level_name, expected_level
):
    """Test that configure_logging respects the STORMI_LOG_LEVEL environment variable."""
    monkeypatch.setenv("STORMI_LOG_LEVEL", level_name)
    logger = configure_logging()
    assert logger.level == expected_level


def test_configure_logging_defaults_to_info_for_invalid_level(monkeypatch):
    """Test that configure_logging defaults to INFO for invalid log levels."""
    monkeypatch.setenv("STORMI_LOG_LEVEL", "INVALID")
    logger = configure_logging()
    assert logger.level == logging.INFO



"""Tests for the _filtering module."""

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData

from stormi.preprocessing import filter_genes, filter_motif_scores, filter_regions


@pytest.fixture
def simple_adata_rna():
    """Create a small RNA AnnData object for testing."""
    n_cells, n_genes = 20, 30
    X = np.random.rand(n_cells, n_genes)
    gene_names = [f"gene{i}" for i in range(n_genes)]
    # Add some mitochondrial genes
    gene_names[0] = "MT-gene1"
    gene_names[1] = "mt-gene2"

    adata = AnnData(X=X)
    adata.var_names = pd.Index(gene_names)
    adata.var["gene_type"] = ["protein_coding"] * n_genes

    return adata


@pytest.fixture
def simple_adata_atac():
    """Create a small ATAC AnnData object for testing."""
    n_cells, n_peaks = 20, 40
    X = np.random.rand(n_cells, n_peaks)
    peak_names = [f"chr1:{i * 1000}-{i * 1000 + 500}" for i in range(n_peaks)]

    adata = AnnData(X=X)
    adata.var_names = pd.Index(peak_names)

    return adata


@pytest.fixture
def simple_motif_scores():
    """Create simple motif scores for testing."""
    genes = [f"gene{i}" for i in range(5)]
    peaks = [f"chr1:{i * 1000}-{i * 1000 + 500}" for i in range(8)]

    data = []
    for gene in genes:
        for peak in peaks:
            data.append({"gene": gene, "peak": peak, "score": np.random.rand()})

    return pd.DataFrame(data)


def test_filter_genes(simple_adata_rna):
    """Test filtering genes from an AnnData object."""
    # Basic filtering
    filtered = filter_genes(simple_adata_rna)

    # Check that mitochondrial genes are removed when requested
    assert filtered.shape[1] < simple_adata_rna.shape[1]
    assert not any(g.startswith(("MT-", "mt-")) for g in filtered.var_names)


def test_filter_regions(simple_adata_atac):
    """Test filtering regions from an AnnData object."""
    # Set some regions to be detected in few cells
    X_sparse = simple_adata_atac.X.copy()
    X_sparse[:, 0] = np.zeros(simple_adata_atac.shape[0])  # No cells have this peak
    X_sparse[:, 1] = np.zeros(simple_adata_atac.shape[0])
    X_sparse[0, 1] = 1  # Only one cell has this peak
    simple_adata_atac.X = X_sparse

    # Filter with min_cells=2 and score_percentile=0.0 (disable score-based filtering)
    filtered = filter_regions(simple_adata_atac, min_cells=2, score_percentile=0.0)

    # Check that only the regions with less than min_cells were filtered
    assert filtered.shape[1] == simple_adata_atac.shape[1] - 2


def test_filter_motif_scores(simple_adata_rna, simple_adata_atac, simple_motif_scores):
    """Test filtering motif scores based on genes and regions."""
    # Create a motif_scores DataFrame in the expected format (with TF names as index)
    # First get a subset of genes and peaks
    rna_genes = simple_adata_rna.var_names[:3]
    atac_peaks = simple_adata_atac.var_names[:3]

    # Subset RNA and ATAC data to these genes and peaks
    subset_rna = simple_adata_rna[:, rna_genes]
    subset_atac = simple_adata_atac[:, atac_peaks]

    # Create motif scores with TFs as index and peaks as columns
    tf_peak_data = {}
    for peak in atac_peaks:
        tf_peak_data[peak] = np.random.rand(len(rna_genes))

    motif_df = pd.DataFrame(tf_peak_data, index=rna_genes)

    # Call the filter_motif_scores function
    filtered_scores = filter_motif_scores(motif_df, subset_rna, subset_atac)

    # Check filtering worked correctly
    assert set(filtered_scores.index) <= set(subset_rna.var_names)
    assert set(filtered_scores.columns) <= set(subset_atac.var_names)



from __future__ import annotations

from collections import OrderedDict
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
import pandas as pd
from anndata import AnnData
from beartype import beartype
from jax import debug, lax
from numpyro.handlers import scale
from scipy import sparse

from .RNA_utils import sample_prior


@jax.jit
def mlp(params: dict[str, jnp.ndarray], x: jnp.ndarray) -> jnp.ndarray:
    """
    Interpretable 2-layer MLP with positive outputs:
      - Hidden modules use ReLU thresholding
      - Linear readout gives gene logits, then softplus ensures positivity

    Args:
        params: dict with 'W1','b1','W2','b2'.
        x:      Array (batch_size, num_tfs) of TF levels.

    Returns:
        Array (batch_size, num_genes) of non-negative transcription rates.
    """
    # Module activations (logical combination of TFs)
    h = jax.nn.elu(jnp.dot(x, params["W1"]) + params["b1"])
    # Linear read-out to genes
    logits = jnp.dot(h, params["W2"]) + params["b2"]
    # Enforce non-negativity: softplus activation
    out = jax.nn.softplus(logits)
    return out


def _linear_dt_grid(n_points: int, ratio: float) -> jnp.ndarray:
    """Return an array `t[0..n_points-1] ∈ [0,1]` where
    dt(i) grows linearly from dt_min to dt_max and dt_max/dt_min = *ratio*.
    The grid size is *n_points* (so dt array has length n_points-1).
    """
    ratio = float(max(ratio, 1.0))  # ensure ≥ 1
    N = n_points - 1  # number of intervals
    idx = jnp.arange(N, dtype=jnp.float32)
    weights = 1.0 + (ratio - 1.0) * idx / (N - 1)  # linear ramp 1 → ratio
    dt_norm = weights / weights.sum()  # normalise so Σ dt = 1
    times = jnp.concatenate([jnp.array([0.0], dtype=jnp.float32), jnp.cumsum(dt_norm)])
    return times


@beartype
def prepare_model_input(
    adata_rna: AnnData,
    tf_list: List[str],
    n_cells_col: str = "n_cells",
    prior_time_col: Optional[str] = None,
    batch_annotation: Optional[str] = None,
    prior_path_col: Optional[str] = None,
    terminal_states: Optional[List[str]] = None,
    initial_states: Optional[List[str]] = None,
    cluster_key: Optional[str] = None,
    prior_timespan: Optional[float] = 40,
    N_steps: int = 100,
    grid_type: str = "linear",
    grid_param: float = 1.0,
    custom_grid_fn: Optional[Callable[[int], jnp.ndarray]] = None,
) -> Dict[str, Any]:
    """
    Prepare RNA input data for the model by extracting spliced and unspliced counts,
    computing transcription factor indices, and extracting cell metadata including batch
    information from adata_rna.obs if provided.  In addition, if you supply
    `cluster_key`, `initial_states`, and `terminal_states`, this will automatically
    build a “Prior Paths” (1-based) from those labels.  If you still want to pass
    in a numeric column (`prior_path_col`) you can; that takes priority if given.

    Parameters
    ----------
    adata_rna : AnnData
        AnnData object containing RNA expression data with 'spliced' and 'unspliced' layers.
    tf_list : List[str]
        List of transcription factor names.
    n_cells_col : str, optional
        Column name in `adata_rna.obs` representing the number of cells per metacell.
    prior_time_col : Optional[str], optional
        Column name in `adata_rna.obs` containing prior pseudotimes; if provided, used
        to center the prior over t_c (default: None).
    batch_annotation : Optional[str], optional
        Column name in `adata_rna.obs` that contains batch information; if provided,
        cells are assigned batch indices based on this column.
    prior_path_col : Optional[str], optional
        (legacy) Column name in `adata_rna.obs` that contains a hard‐coded 1‐based path index;
        missing/NaN entries become −1 (inference).  Takes priority over the
        automatic cluster‐based labeling described below.
    terminal_states : Optional[List[str]], optional
        If provided (and `cluster_key` is also given), defines a list of cluster names
        that should be mapped to terminal paths.  They will be assigned raw_prior = 1..P
        in the order you list them.
    initial_states : Optional[List[str]], optional
        If provided (and `cluster_key` is also given), defines a list of cluster names
        that should be mapped to raw_prior = 0 (which then becomes prior_path = −2
        internally, i.e. “initial state”).
    cluster_key : Optional[str], optional
        Column in `adata_rna.obs` whose values are compared against `initial_states`
        or `terminal_states` to build a “Prior Path” assignment.  This is only used
        if `prior_path_col` is None.
    prior_timespan : float, optional
        Used if prior_time_col is None (default span).  Otherwise recalculated from data.
    N_steps : int, optional
        Number of grid steps (default: 100).
    grid_type : str, optional
        “linear” or “linear_dt” (default: “linear”).
    grid_param : float, optional
        Parameter for “linear_dt” spacing (ignored if grid_type=“linear”).
    custom_grid_fn : Optional[Callable[[int], jnp.ndarray]], optional
        If you want to supply your own function to build `times_norm`.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing:
          - `data`: JAX array of shape (cells, genes, 2) with stacked unspliced and spliced counts.
          - `M_c`: JAX array of shape (cells, 1, 1) holding the per‐cell metacell size.
          - `obs2sample`: one‐hot encoding of batch indices (shape cells × n_batch).
          - `batch_index`: JAX array of shape (cells,) with batch indices.
          - `tf_indices`: JAX array of TF gene indices.
          - `total_num_cells`: int, number of cells.
          - `n_batch`: int, number of batches.
          - `prior_time`: JAX array of prior times or None.
          - `prior_timespan`: float, span of prior_time or default.
          - `prior_path`: JAX int array of shape (cells,) with −1 or 0‐based path ids;
              prior_path = −2 means “initial state,” prior_path ≥ 0 means fixed to path id.
          - `unknown_idx`: Python list of ints, indices of cells with unknown prior (prior_path < 0).
          - `unknown_count`: int, number of unknown cells.
          - `times_norm`, `dt_norm`, `T_limits`, `hidden_units`, `num_paths`
    """

    def _to_dense_f32(layer):
        if sparse.issparse(layer):
            # cast to float32 and densify in one go
            return layer.astype(np.float32).toarray()
        else:
            # if it’s already an ndarray, just ensure dtype
            return np.asarray(layer, dtype=np.float32)

    # ========== 1) STACK RAW DATA ==========
    spliced = _to_dense_f32(adata_rna.layers["spliced"])
    unspliced = _to_dense_f32(adata_rna.layers["unspliced"])
    data = jnp.array(np.stack([unspliced, spliced], axis=-1))

    # ========== 2) METACELL SIZES ==========
    M_c = jnp.array(
        np.expand_dims(np.expand_dims(adata_rna.obs[n_cells_col].to_numpy(), -1), -1),
        dtype=jnp.float32,
    )
    total_num_cells = int(data.shape[0])

    # ========== 3) PRIOR TIME ==========
    if prior_time_col is not None:
        prior_time = jnp.array(
            adata_rna.obs[prior_time_col].to_numpy(), dtype=jnp.float32
        )
        prior_timespan = float(prior_time.max() - prior_time.min())
    else:
        prior_time = None

    # ========== 4) BATCH INDEX ==========
    if batch_annotation is not None:
        batches = adata_rna.obs[batch_annotation].astype(str)
        uniq = batches.unique()
        mapping = {b: i for i, b in enumerate(uniq)}
        batch_index = jnp.array(batches.map(mapping).to_numpy(), dtype=jnp.int32)
    else:
        batch_index = jnp.zeros(total_num_cells, dtype=jnp.int32)
    n_batch = int(jnp.unique(batch_index).shape[0])
    obs2sample = jax.nn.one_hot(batch_index, num_classes=n_batch, dtype=jnp.float32)

    # ========== 5) TF INDICES ==========
    is_tf = np.isin(adata_rna.var_names, tf_list)
    tf_indices = jnp.array(np.where(is_tf)[0], dtype=jnp.int32)
    hidden_units = int(jnp.sqrt(int(np.sum(is_tf))))

    # ========== 6) PRIOR PATH (either from prior_path_col OR from clusters) ==========
    if prior_path_col is not None:
        raw_prior = adata_rna.obs[prior_path_col].to_numpy()
    else:
        if (
            (cluster_key is None)
            or (initial_states is None)
            or (terminal_states is None)
        ):
            raise ValueError(
                "Either supply `prior_path_col`, or else supply `cluster_key`, "
                "`initial_states`, and `terminal_states` so that the function can build a prior."
            )

        # Initialize everything to NA (we will treat NA as “unknown” => prior_path = −1).
        raw_prior = pd.Series(
            data=[pd.NA] * total_num_cells, index=adata_rna.obs.index, dtype="object"
        )

        # 6a) Assign initial states → raw_prior = 0
        init_mask = adata_rna.obs[cluster_key].isin(initial_states)
        raw_prior.loc[init_mask] = 0

        # 6b) Assign terminal states → raw_prior = 1,2,… in the order given
        for idx, term_label in enumerate(terminal_states, start=1):
            mask = adata_rna.obs[cluster_key] == term_label
            raw_prior.loc[mask] = idx
        raw_prior = raw_prior.to_numpy()

    valid_paths = [int(v) for v in raw_prior if pd.notna(v) and int(v) > 0]
    num_paths = max(valid_paths) if valid_paths else 1

    prior_path_np = np.full((total_num_cells,), -1, dtype=np.int32)  # −1 ⇒ unknown
    for i, v in enumerate(raw_prior):
        if pd.isna(v):
            # leave as −1
            continue
        v_int = int(v)
        if v_int == 0:
            # “initial state” → prior_path = −2
            prior_path_np[i] = -2
        else:
            # 1…P → fixed to path v−1
            prior_path_np[i] = v_int - 1

    prior_path = jnp.array(prior_path_np, dtype=jnp.int32)
    unknown_idx = [int(i) for i, p in enumerate(prior_path_np) if p < 0]
    unknown_count = len(unknown_idx)

    # ========== 7) BUILD TIMEGRID ==========
    if prior_time_col is None:
        T_min = -prior_timespan / 2.0
        T_max = +prior_timespan / 2.0
    else:
        T_min = prior_time.min() - prior_timespan / 6.0
        T_max = prior_time.max() + prior_timespan / 6.0

    if grid_type == "linear":
        times_norm = jnp.linspace(0.0, 1.0, N_steps + 1)
    elif grid_type in {"linear_dt", "lin_dt"}:
        times_norm = _linear_dt_grid(N_steps + 1, ratio=float(grid_param))
    else:
        raise ValueError(f"Unknown grid_type '{grid_type}'.")

    dt_norm = jnp.diff(times_norm)  # length = N_steps

    return OrderedDict(
        [
            ("data", data),
            ("M_c", M_c),
            ("obs2sample", obs2sample),
            ("batch_index", batch_index),
            ("tf_indices", tf_indices),
            ("total_num_cells", total_num_cells),
            ("n_batch", n_batch),
            ("prior_time", prior_time),
            ("prior_timespan", prior_timespan),
            ("prior_path", prior_path),
            ("unknown_idx", unknown_idx),
            ("unknown_count", unknown_count),
            ("times_norm", times_norm),
            ("dt_norm", dt_norm),
            ("T_limits", (T_min, T_max)),
            ("hidden_units", hidden_units),
            ("num_paths", num_paths),
        ]
    )


def soft_clip(y, lower, upper, sharpness=1.0):
    y1 = lower + jax.nn.softplus(sharpness * (y - lower)) / sharpness
    y2 = upper - jax.nn.softplus(sharpness * (upper - y1)) / sharpness
    return y2


jax.jit


def euler_maruyama(
    times_grid: jnp.ndarray,
    dt: jnp.ndarray,
    y0: jnp.ndarray,
    drift_fn,
    diffusion_fn,
    drift_args,
    diff_args,
    eps_grid: jnp.ndarray,
) -> jnp.ndarray:
    """
    Perform Euler–Maruyama integration on a fixed time grid for multiple sample paths.

    Parameters:
    - times_grid (jnp.ndarray): 1D array of time points of length L.
    - dt (jnp.ndarray): 1D array of time step sizes between grid points, length L-1.
    - y0 (jnp.ndarray): Initial state for each path, shape (num_paths, num_genes, 3).
    - drift_fn (Callable): Function computing deterministic drift: f = drift_fn(t, y, drift_args).
    - diffusion_fn (Callable): Function computing stochastic diffusion: g = diffusion_fn(t, y, diff_args).
    - drift_args (Any): Extra arguments to pass to drift_fn.
    - diff_args (Any): Extra arguments to pass to diffusion_fn.
    - eps_grid (jnp.ndarray): Pre-sampled Gaussian noise, shape (num_paths, L-1, num_genes).

    Returns:
    - jnp.ndarray: Simulated state trajectory, shape (L, num_paths, num_genes, 3).
    """
    eps = jnp.swapaxes(eps_grid, 0, 1)

    def body_fn(carry, inputs):
        y_prev, t_prev = carry
        eps_i, dt_i = inputs
        f = drift_fn(t_prev, y_prev, drift_args) * dt_i
        g = diffusion_fn(t_prev, y_prev, diff_args)
        st = g * (jnp.sqrt(dt_i)[None, None, None] * eps_i[..., None])
        y_new = y_prev + f + st
        y_new = jnp.clip(y_new, 0.0, 1000)
        return (y_new, t_prev + dt_i), y_new

    (y_last, _), ys = jax.lax.scan(body_fn, (y0, times_grid[0]), (eps, dt))
    return jnp.concatenate([y0[None], ys], axis=0)


@jax.jit
def interpolate_solution(
    sol_grid: jnp.ndarray,
    times_grid: jnp.ndarray,
    T_c: jnp.ndarray,
) -> jnp.ndarray:
    """
    Linearly interpolate a precomputed solution grid at specified cell times.

    Parameters:
    - sol_grid (jnp.ndarray): Solution values on a regular time grid, shape (L, num_paths, num_genes, 3).
    - times_grid (jnp.ndarray): 1D array of time points corresponding to sol_grid, length L.
    - T_c (jnp.ndarray): Array of arbitrary times at which to interpolate, shape (num_cells,).

    Returns:
    - jnp.ndarray: Interpolated solution at each time in T_c, shape (num_cells, num_paths, num_genes, 3).
    """
    idx0 = jnp.searchsorted(times_grid, T_c, side="right") - 1
    idx0 = jnp.clip(idx0, 0, sol_grid.shape[0] - 2)
    t0 = times_grid[idx0]
    t1 = times_grid[idx0 + 1]
    sol0 = sol_grid[idx0]
    sol1 = sol_grid[idx0 + 1]
    w = ((T_c - t0) / (t1 - t0))[..., None, None, None]
    return sol0 * (1 - w) + sol1 * w


@jax.jit
def diffusion_fn(t, y, diff_args):
    """
    Compute the diffusion term for the SDE, nonzero only on the protein dimension
    for specified transcription factor genes.

    Parameters:
    - t (float): Current time.
    - y (jnp.ndarray): Current state array, shape (num_paths, num_genes, 3).
    - diff_args (tuple): Tuple containing:
        - sigma_tf (jnp.ndarray): Diffusion strengths for each TF, shape (num_tfs,).
        - tf_indices (array-like): Indices of transcription factor genes.

    Returns:
    - jnp.ndarray: Diffusion contribution, same shape as y.
    """
    sigma_tf, tf_indices = diff_args
    return jnp.zeros_like(y).at[:, tf_indices, 2].set(sigma_tf[None, :])


@jax.jit
def drift_fn(t, state, args):
    """
    Compute the deterministic drift for the coupled transcription–splicing–protein system.

    Parameters:
    - t (float): Current time.
    - state (jnp.ndarray): Current state, shape (num_paths, num_genes, 3), containing
      unspliced (u), spliced (s), and protein (p) counts.
    - args (tuple): Tuple containing:
        - alpha_0 (jnp.ndarray): Baseline transcription rates, shape (num_genes,).
        - beta_g (jnp.ndarray): Splicing rates per gene, shape (num_genes,).
        - gamma_g (jnp.ndarray): Degradation rates per gene, shape (num_genes,).
        - lamda (jnp.ndarray): Translation rates for TF proteins, shape (num_tfs,).
        - kappa (jnp.ndarray): Protein degradation rates for TFs, shape (num_tfs,).
        - nn_params (dict): Neural network parameters for computing regulation.
        - tf_indices (array-like): Indices of transcription factor genes.
        - T_ON (float): Switch-on time threshold.

    Returns:
    - jnp.ndarray: Time derivative of the state, same shape as state.
    """
    alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices = args

    u = jnp.clip(state[..., 0], 0, 10 ** (3))
    s = jnp.clip(state[..., 1], 0, 10 ** (3))
    p = jnp.clip(state[..., 2], 0, 10 ** (3))

    alpha = alpha_0 * jnp.clip(mlp(nn_params, p[:, tf_indices]), 0, 10 ** (3))
    du_dt = alpha - beta_g * u
    ds_dt = beta_g * u - gamma_g * s

    dp_dt = jnp.zeros_like(u)
    dp_dt = dp_dt.at[:, tf_indices].set(
        lamda * s[:, tf_indices] - kappa * p[:, tf_indices]
    )

    return jnp.stack([du_dt, ds_dt, dp_dt], axis=-1)

# Define the complete NumPyro model
def model(
    data: Any,
    M_c: Any,
    obs2sample: Any,
    batch_index: Any,
    tf_indices: Any,
    total_num_cells: int,
    n_batch: int,
    prior_time: Any,
    prior_timespan: Any,
    prior_path: Any,
    unknown_idx: Any,
    unknown_count: int,
    times_norm: Any,
    dt_norm: Any,
    T_limits: Any,
    hidden_units: int,
    num_paths: int,
    return_alpha: bool = False,
    Tmax_alpha: float = 50.0,
    Tmax_beta: float = 1.0,
    splicing_rate_alpha_hyp_prior_alpha: float = 20.0,
    splicing_rate_alpha_hyp_prior_mean: float = 5.0,
    splicing_rate_mean_hyp_prior_alpha: float = 10.0,
    splicing_rate_mean_hyp_prior_mean: float = 1.0,
    degradation_rate_alpha_hyp_prior_alpha: float = 20.0,
    degradation_rate_alpha_hyp_prior_mean: float = 5.0,
    degradation_rate_mean_hyp_prior_alpha: float = 10.0,
    degradation_rate_mean_hyp_prior_mean: float = 1.0,
    transcription_rate_alpha_hyp_prior_alpha: float = 20.0,
    transcription_rate_alpha_hyp_prior_mean: float = 2.0,
    transcription_rate_mean_hyp_prior_alpha: float = 10.0,
    transcription_rate_mean_hyp_prior_mean: float = 5.0,
    lambda_alpha: float = 1.0,
    lambda_mean: float = 1.0,
    kappa_alpha: float = 1.0,
    kappa_mean: float = 1.0,
    detection_mean_hyp_prior_alpha: float = 1.0,
    detection_mean_hyp_prior_beta: float = 1.0,
    detection_hyp_prior_alpha: float = 10.0,
    detection_i_prior_alpha: float = 100.0,
    detection_gi_prior_alpha: float = 200.0,
    gene_add_alpha_hyp_prior_alpha: float = 9.0,
    gene_add_alpha_hyp_prior_beta: float = 3.0,
    gene_add_mean_hyp_prior_alpha: float = 1.0,
    gene_add_mean_hyp_prior_beta: float = 100.0,
    stochastic_v_ag_hyp_prior_alpha: float = 9.0,
    stochastic_v_ag_hyp_prior_beta: float = 3.0,
    init_seed: int = 0,
):
    """
    NumPyro model for coupled transcription and splicing dynamics.

    Args:
        data: Observed data array of shape (num_cells, num_genes, num_modalities).
        M_c: Number of cells in each metacell.
        batch_index: Array indicating batch assignments for each cell.
        tf_indices: Indices of genes that are TFs.
        total_num_cells: Number of cells in the full dataset.
        n_batch: Number of batches.
        Tmax_alpha: Alpha parameter for Tmax prior.
        Tmax_beta: Beta parameter for Tmax prior.
        ... (other hyperparameters for priors)
        key: Random number generator key.

    Returns:
        None. Defines the probabilistic model for inference.
    """

    num_cells = int(data.shape[0])
    num_genes = int(data.shape[1])
    num_modalities = int(data.shape[2])
    num_tfs = int(tf_indices.shape[0])
    batch_size = num_cells
    n_keys = 2
    base_key = jax.random.PRNGKey(init_seed)
    all_keys = jax.random.split(base_key, n_keys)

    # Splicing Rates for mRNA
    splicing_alpha = numpyro.sample(
        "splicing_alpha",
        dist.Gamma(
            splicing_rate_alpha_hyp_prior_alpha,
            splicing_rate_alpha_hyp_prior_alpha / splicing_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_alpha = jnp.clip(splicing_alpha, a_min=1e-2, a_max=1e3)

    splicing_mean = numpyro.sample(
        "splicing_mean",
        dist.Gamma(
            splicing_rate_mean_hyp_prior_alpha,
            splicing_rate_mean_hyp_prior_alpha / splicing_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_mean = jnp.clip(splicing_mean, a_min=1e-2, a_max=1e3)

    beta_g = numpyro.sample(
        "beta_g",
        dist.Gamma(splicing_alpha, splicing_alpha / splicing_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )

    # Degradation Rates for mRNA
    degradation_alpha = numpyro.sample(
        "degradation_alpha",
        dist.Gamma(
            degradation_rate_alpha_hyp_prior_alpha,
            degradation_rate_alpha_hyp_prior_alpha
            / degradation_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_alpha = degradation_alpha + 0.001  # Prevent zero

    degradation_mean = numpyro.sample(
        "degradation_mean",
        dist.Gamma(
            degradation_rate_mean_hyp_prior_alpha,
            degradation_rate_mean_hyp_prior_alpha
            / degradation_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_mean = jnp.clip(degradation_mean, a_min=1e-2, a_max=1e3)

    gamma_g = numpyro.sample(
        "gamma_g",
        dist.Gamma(
            degradation_alpha, degradation_alpha / degradation_mean, validate_args=True
        )
        .expand([num_genes])
        .to_event(1),
    )

    # Translation rate for proteins
    lamda = numpyro.sample(
        "lambda",
        dist.Gamma(lambda_alpha, lambda_alpha / lambda_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )
    # Degradation rate for proteins
    kappa = numpyro.sample(
        "kappa",
        dist.Gamma(kappa_alpha, kappa_alpha / kappa_mean, validate_args=True)
        .expand([num_tfs])
        .to_event(1),
    )

    # Time Parameters
    # 1) sample a *single* global scaling factor
    T_min, T_max = T_limits
    T_scaling = numpyro.sample("T_scaling", dist.Beta(1.0, 3.0))
    total_span = (T_max - T_min) * T_scaling

    # 2) build *static‐shape* grid (always N_steps long)
    times_grid = T_min + times_norm * total_span
    dt = dt_norm * total_span

    # convert your absolute prior_time (if any) into a [0,1] loc/scale
    if prior_time is None:
        prior_time = T_min + (T_max - T_min) / 2
        prior_time_sd = prior_timespan / 2.0
    else:
        prior_time_sd = prior_timespan / 12.0

    with numpyro.plate("cells", batch_size):
        t_c = numpyro.sample(
            "t_c",
            dist.TruncatedNormal(
                loc=prior_time,
                scale=prior_time_sd,
                low=T_limits[0],
                high=T_limits[1],
            ),
        )
        T_c = numpyro.deterministic("T_c", T_scaling * t_c)

    # ============= Expression model =============== #

    # scale of alpha
    alpha_0 = numpyro.sample(
        "alpha_0",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([num_genes]).to_event(1),
    )

    # Initial Conditions for ODE, sampling only for spliced and unspliced
    initial_state_2d = numpyro.sample(
        "initial_state_2d",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([1, num_genes, 2]).to_event(1),
    )

    # Create the third dimension containing proteins, only for TFs:
    third_dimension = jnp.zeros((1, num_genes))
    third_dimension = third_dimension.at[0, tf_indices].set(
        numpyro.sample(
            "initial_state_tf",
            dist.Gamma(1.0, 1.0, validate_args=True).expand([len(tf_indices)]),
        )
    )

    # Concatenate the dimensions to form the final initial state
    initial_state = jnp.concatenate(
        [initial_state_2d, third_dimension[..., None]], axis=-1
    )

    # Neural Net params
    H = hidden_units
    W1 = numpyro.param(
        "W1", jax.random.normal(all_keys[0], (tf_indices.shape[0], H)) * 0.01
    )
    b1 = numpyro.param("b1", jnp.zeros((H,)))
    W2 = numpyro.param(
        "W2", jax.random.normal(all_keys[1], (H, data.shape[1])) * 0.01
    )
    b2 = numpyro.param("b2", jnp.zeros((data.shape[1],)))
    nn_params = {"W1": W1, "b1": b1, "W2": W2, "b2": b2}

    # # ---- hierarchical log-σ prior --------------------
    log_tau = numpyro.sample("log_tau_sigma", dist.Normal(jnp.log(0.0005), 0.0005))

    log_lam = numpyro.sample(
        "log_lam_sigma", dist.Normal(0.0, 0.001).expand([num_tfs]).to_event(1)  # ← here
    )

    log_sigma_tf = log_tau + log_lam
    sigma_tf = soft_clip(jnp.exp(log_sigma_tf), 10 ** (-12), 10 ** (-3))

    ode_args = (alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices)
    diff_args = (sigma_tf, tf_indices)

    # ───────── Build three masks ─────────
    known_mask = prior_path >= 0  # fixed-path cells
    initial_mask = prior_path == -2  # “initial-state” cells

    one_hot_known = jax.nn.one_hot(jnp.clip(prior_path, 0, num_paths - 1), num_paths)

    # ───────── Hyperparameters for Logistic-Normal ─────────
    sigma_init = 0.005
    sigma_unknown = 1.0

    sigma_matrix = jnp.where(
        initial_mask[:, None],
        jnp.ones((1, num_paths)) * sigma_init,
        jnp.ones((1, num_paths)) * sigma_unknown,
    )

    with numpyro.plate("cells", batch_size):
        z_pw = numpyro.sample(
            "z_pw",
            dist.Normal(loc=jnp.zeros((1, num_paths)), scale=sigma_matrix).to_event(1),
        )

    w_full = jax.nn.softmax(z_pw, axis=-1)  # (n_cells, num_paths)
    w_final = jnp.where(known_mask[:, None], one_hot_known, w_full)

    path_weights = numpyro.deterministic("path_weights", w_final)

    path_counts = jnp.sum(path_weights, axis=0)

    # with scale(scale=path_counts[:, None, None]):
    #     eps_grid = numpyro.sample(
    #         "eps_grid",
    #         dist.Normal(0.0, 1.0)
    #         .expand([num_paths, dt.shape[0], num_genes])
    #         .to_event(3),
    #     )

    # ── Correct way to draw independent noise for each of the num_paths ──
    with scale(scale=path_counts[:, None, None]):
        with numpyro.plate("paths", num_paths):
            eps_grid = numpyro.sample(
                "eps_grid",
                dist.Normal(0.0, 1.0)
                    .expand([dt.shape[0], num_genes])
                    .to_event(2),
            )

    # integrate once on the grid
    sol_grid = euler_maruyama(
        times_grid,
        dt,
        jnp.tile(initial_state, (num_paths, 1, 1)),
        drift_fn,
        diffusion_fn,
        ode_args,
        diff_args,
        eps_grid,
    )

    # interpolate each cell at its own time T_c
    sol_at_cells = numpyro.deterministic(
        "sol_at_cells", interpolate_solution(sol_grid, times_grid, T_c)
    )

    # weighted average over paths
    weighted_preds = numpyro.deterministic(
        "predictions_rearranged",
        jnp.sum(sol_at_cells * path_weights[:, :, None, None], axis=1),
    )

    # use weighted_preds[..., :2] as mu_expression downstream
    mu_expression = jnp.clip(weighted_preds[..., :2], a_min=1e-5, a_max=1e5)

    if return_alpha:

        # 1) Extract full protein at each grid point:
        P_full = sol_grid[..., 2]  # shape = (L, P, G)

        # 2) Compute “protein before the noise” for each step t=1..L-1:
        # t_vec has shape (L-1,) and sol_grid[:-1] has shape (L-1, P, G, 3).
        dp_dt_full = jax.vmap(
            lambda t, y: drift_fn(t, y, ode_args),
            in_axes=(0, 0)
        )(times_grid[:-1], sol_grid[:-1])  # → shape (L-1, P, G, 3)
        dp_dt_grid = dp_dt_full[..., 2]  # shape (L-1, P, G)

        det_increment = dp_dt_grid * dt[:, None, None]
        P_before_noise = sol_grid[:-1, ..., 2] + det_increment                     # (L-1, P, G)

        # 3) Select TF‐protein columns and flatten to feed into mlp:
        tf_prot_full   = P_full[1:, :, tf_indices]         # (L-1, P, num_tfs)
        tf_prot_before = P_before_noise[:, :, tf_indices]  # (L-1, P, num_tfs)
        flat_full      = tf_prot_full.reshape(-1, len(tf_indices))    # ((L-1)*P, num_tfs)
        flat_before    = tf_prot_before.reshape(-1, len(tf_indices))  # ((L-1)*P, num_tfs)

        # 4) Compute α for each gene at each (t, p):
        out_full_flat   = mlp(nn_params, flat_full)    # ((L-1)*P, num_genes)
        out_before_flat = mlp(nn_params, flat_before)  # ((L-1)*P, num_genes)

        # multiply by alpha_0 and reshape to (L-1, P, G)
        alpha_full = (alpha_0 * out_full_flat).reshape(tf_prot_full.shape[0],
                                                       tf_prot_full.shape[1],
                                                       alpha_0.shape[0])
        alpha_before = (alpha_0 * out_before_flat).reshape(tf_prot_before.shape[0],
                                                           tf_prot_before.shape[1],
                                                           alpha_0.shape[0])

        numpyro.deterministic("alpha_full_grid",   alpha_full)   # (L-1, P, G)
        numpyro.deterministic("alpha_before_grid", alpha_before) # (L-1, P, G)

    # ───── Measurement noise ─────

    detection_mean_y_e = numpyro.sample(
        "detection_mean_y_e",
        dist.Beta(
            jnp.ones((1, 1)) * detection_mean_hyp_prior_alpha,
            jnp.ones((1, 1)) * detection_mean_hyp_prior_beta,
            validate_args=True,
        )
        .expand([n_batch, 1])
        .to_event(2),
    )

    beta = ((detection_hyp_prior_alpha / (obs2sample @ detection_mean_y_e)).T)[0, ...]

    with numpyro.plate("cells", batch_size):
        detection_y_c = numpyro.sample(
            "detection_y_c", dist.Gamma(detection_hyp_prior_alpha, beta)
        )

    detection_y_i = numpyro.sample(
        "detection_y_i",
        dist.Gamma(detection_i_prior_alpha, detection_i_prior_alpha, validate_args=True)
        .expand([1, 1, 2])
        .to_event(3),
    )

    detection_y_gi = numpyro.sample(
        "detection_y_gi",
        dist.Gamma(
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            jnp.ones((1, 1)) * detection_gi_prior_alpha,
            validate_args=True,
        )
        .expand([1, num_genes, 2])
        .to_event(3),
    )

    # ======= Gene-specific additive component (Ambient RNA/ "Soup") for spliced and unspliced counts ====== #

    s_g_gene_add_alpha_hyp = numpyro.sample(
        "s_g_gene_add_alpha_hyp",
        dist.Gamma(
            gene_add_alpha_hyp_prior_alpha,
            gene_add_alpha_hyp_prior_beta,
            validate_args=True,
        ),
        sample_shape=(2,),
    )

    s_g_gene_add_mean = numpyro.sample(
        "s_g_gene_add_mean",
        dist.Gamma(
            gene_add_mean_hyp_prior_alpha,
            gene_add_mean_hyp_prior_beta,
            validate_args=True,
        )
        .expand([n_batch, 1, 2])
        .to_event(3),
    )

    s_g_gene_add_alpha_e_inv = numpyro.sample(
        "s_g_gene_add_alpha_e_inv",
        dist.Exponential(s_g_gene_add_alpha_hyp).expand([n_batch, 1, 2]).to_event(3),
    )

    s_g_gene_add_alpha_e = jnp.ones((1, 1)) / s_g_gene_add_alpha_e_inv**2

    s_g_gene_add = numpyro.sample(
        "s_g_gene_add",
        dist.Gamma(
            s_g_gene_add_alpha_e,
            s_g_gene_add_alpha_e / s_g_gene_add_mean,
            validate_args=True,
        )
        .expand([n_batch, num_genes, 2])
        .to_event(3),
    )

    # ========= Gene-specific overdispersion of spliced and unspliced counts ============== #

    stochastic_v_ag_hyp = numpyro.sample(
        "stochastic_v_ag_hyp",
        dist.Gamma(
            stochastic_v_ag_hyp_prior_alpha,
            stochastic_v_ag_hyp_prior_beta,
            validate_args=True,
        )
        .expand([1, 2])
        .to_event(2),
    )

    stochastic_v_ag_inv = numpyro.sample(
        "stochastic_v_ag_inv",
        dist.Exponential(stochastic_v_ag_hyp).expand([1, num_genes, 2]).to_event(3),
    )

    stochastic_v_ag = jnp.ones((1, 1)) / stochastic_v_ag_inv**2

    # ===================== Expected expression ======================= #

    additive_term = numpyro.deterministic(
        "additive_term", jnp.einsum("cb,bgi->cgi", obs2sample, s_g_gene_add)
    )

    normalizing_term = numpyro.deterministic(
        "normalizing_term",
        detection_y_c[:, None, None] * detection_y_i * detection_y_gi * M_c,
    )

    mu = numpyro.deterministic("mu", (mu_expression + additive_term) * normalizing_term)

    # Data likelihood
    concentration = stochastic_v_ag * M_c
    rate = concentration / mu
    scale_factor = total_num_cells / batch_size
    with scale(scale=scale_factor):
        numpyro.sample("data_target", dist.GammaPoisson(concentration, rate), obs=data)



"""
Stormi

Stormi is a computational framework for probabilistic modeling of cellular dynamics,
particularly focusing on RNA transcription, protein expression, and gene regulation.

docs: https://stormi.celldynamix.net
source: https://github.com/celldynamics/stormi
PyPI package: https://pypi.org/project/stormi

Note: preprocessing functionality is available as an optional extra:
- Install it with: pip install stormi[preprocessing]
- Import from it with: from stormi.preprocessing import ...
"""

from importlib import metadata

from stormi import (
    guides,
    logging,
    models,
    plots,
    posterior,
    styles,
    train,
)
from stormi.main import main

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    __version__ = "unknown"

del metadata

__all__ = [
    "guides",
    "logging",
    "models",
    "plots",
    "posterior",
    "preprocessing",
    "styles",
    "train",
    "main",
]



"""Stormi preprocessing subpackage for processing RNA and ATAC-seq data.

This module provides functions for preprocessing single-cell RNA and ATAC-seq data,
including filtering, dimensionality reduction, metacell computation, and regulatory
element analysis.

Note: This module requires additional dependencies that are not installed by default.
Install them with: pip install stormi[preprocessing]
"""

import importlib
import logging

logger = logging.getLogger(__name__)

HAS_PREPROCESSING_DEPENDENCIES = True
MISSING_DEPENDENCIES = []


def _check_dependency(module_name: str) -> bool:
    """Check if a dependency is available."""
    global HAS_PREPROCESSING_DEPENDENCIES, MISSING_DEPENDENCIES
    try:
        importlib.import_module(module_name)
        return True
    except ImportError:
        HAS_PREPROCESSING_DEPENDENCIES = False
        MISSING_DEPENDENCIES.append(module_name)
        return False


_has_pyranges = _check_dependency("pyranges")
_has_pybiomart = _check_dependency("pybiomart")
_has_tangermeme = _check_dependency("tangermeme")
_has_gtfparse = _check_dependency("gtfparse")

if not HAS_PREPROCESSING_DEPENDENCIES:
    deps_str = ", ".join(MISSING_DEPENDENCIES)
    logger.warning(
        f"Some preprocessing dependencies are not available: {deps_str}. "
        "Some preprocessing functions will not work. "
        "Install the missing dependencies with: pip install stormi[preprocessing]"
    )

__all__ = [
    # Complexity module
    "compute_pca_complexity",
    "default_n_latent",
    "default_n_hidden",
    "default_n_layers",
    "default_epochs",
    # Embedding module
    "run_scvi",
    # Metacells module
    "compute_metacells",
    "convert_to_dense",
    "create_metacells",
    # Filtering module
    "filter_genes_simple",
    "filter_genes",
    "filter_regions",
    "filter_motif_scores",
    "func_mitochondrial_genes",
    "func_protein_coding_genes",
    "TF_HVG_selection",
    "select_highly_variable_peaks_by_std",
    "keep_promoters_and_select_hv_peaks",
    "bed_file_intersection",
    "simple_bed_intersection",
    # Motifs module
    "compute_motif_scores",
    "compute_in_silico_chipseq",
    # File utils module
    "create_dir_if_not_exists",
    "download_file",
    "unzip_gz",
    "resolve_genome_urls",
    "download_genome_references",
    "check_command_availability",
    # Region-gene-TF module
    "extract_region_tf_pairs",
    "build_gene_tss_dict",
    "parse_region_name",
    "build_pyranges_for_regions",
    "build_pyranges_for_genes",
    "build_region_gene_pairs",
    "construct_region_tf_gene_triplets",
    "rhg_to_rh_indexing",
    # Pipeline module
    "make_multiomic_dataset",
    "pipeline_setup_multiview",
    "preprocessing_pipeline",
]

if HAS_PREPROCESSING_DEPENDENCIES:
    from stormi.preprocessing._complexity import (
        compute_pca_complexity,
        default_epochs,
        default_n_hidden,
        default_n_latent,
        default_n_layers,
    )
    from stormi.preprocessing._embedding import run_scvi
    from stormi.preprocessing._file_utils import (
        check_command_availability,
        create_dir_if_not_exists,
        download_file,
        download_genome_references,
        resolve_genome_urls,
        unzip_gz,
    )
    from stormi.preprocessing._filtering import (
        TF_HVG_selection,
        bed_file_intersection,
        filter_genes,
        filter_genes_simple,
        filter_motif_scores,
        filter_regions,
        func_mitochondrial_genes,
        func_protein_coding_genes,
        keep_promoters_and_select_hv_peaks,
        select_highly_variable_peaks_by_std,
        simple_bed_intersection,
    )
    from stormi.preprocessing._metacells import (
        compute_metacells,
        convert_to_dense,
        create_metacells,
    )
    from stormi.preprocessing._motifs import (
        compute_in_silico_chipseq,
        compute_motif_scores,
    )
    from stormi.preprocessing._pipeline import (
        preprocessing_pipeline,
    )
    from stormi.preprocessing._region_gene_tf import (
        build_gene_tss_dict,
        build_pyranges_for_genes,
        build_pyranges_for_regions,
        build_region_gene_pairs,
        construct_region_tf_gene_triplets,
        extract_region_tf_pairs,
        parse_region_name,
        rhg_to_rh_indexing,
    )
else:

    def _missing_dependencies_error(func_name):
        deps_str = ", ".join(MISSING_DEPENDENCIES)
        raise ImportError(
            f"Cannot use {func_name} because some preprocessing dependencies are missing: {deps_str}. "
            "Install the missing dependencies with: pip install stormi[preprocessing]"
        )

    def compute_pca_complexity(*args, **kwargs):
        _missing_dependencies_error("compute_pca_complexity")

    def default_n_latent(*args, **kwargs):
        _missing_dependencies_error("default_n_latent")

    def default_n_hidden(*args, **kwargs):
        _missing_dependencies_error("default_n_hidden")

    def default_n_layers(*args, **kwargs):
        _missing_dependencies_error("default_n_layers")

    def default_epochs(*args, **kwargs):
        _missing_dependencies_error("default_epochs")

    def run_scvi(*args, **kwargs):
        _missing_dependencies_error("run_scvi")

    def compute_metacells(*args, **kwargs):
        _missing_dependencies_error("compute_metacells")

    def convert_to_dense(*args, **kwargs):
        _missing_dependencies_error("convert_to_dense")

    def create_metacells(*args, **kwargs):
        _missing_dependencies_error("create_metacells")

    def filter_genes(*args, **kwargs):
        _missing_dependencies_error("filter_genes")

    def filter_regions(*args, **kwargs):
        _missing_dependencies_error("filter_regions")

    def filter_motif_scores(*args, **kwargs):
        _missing_dependencies_error("filter_motif_scores")

    def func_mitochondrial_genes(*args, **kwargs):
        _missing_dependencies_error("func_mitochondrial_genes")

    def func_protein_coding_genes(*args, **kwargs):
        _missing_dependencies_error("func_protein_coding_genes")

    def TF_HVG_selection(*args, **kwargs):
        _missing_dependencies_error("TF_HVG_selection")

    def select_highly_variable_peaks_by_std(*args, **kwargs):
        _missing_dependencies_error("select_highly_variable_peaks_by_std")

    def keep_promoters_and_select_hv_peaks(*args, **kwargs):
        _missing_dependencies_error("keep_promoters_and_select_hv_peaks")

    def bed_file_intersection(*args, **kwargs):
        _missing_dependencies_error("bed_file_intersection")

    def simple_bed_intersection(*args, **kwargs):
        _missing_dependencies_error("simple_bed_intersection")

    def compute_motif_scores(*args, **kwargs):
        _missing_dependencies_error("compute_motif_scores")

    def compute_in_silico_chipseq(*args, **kwargs):
        _missing_dependencies_error("compute_in_silico_chipseq")

    def create_dir_if_not_exists(*args, **kwargs):
        _missing_dependencies_error("create_dir_if_not_exists")

    def download_file(*args, **kwargs):
        _missing_dependencies_error("download_file")

    def unzip_gz(*args, **kwargs):
        _missing_dependencies_error("unzip_gz")

    def resolve_genome_urls(*args, **kwargs):
        _missing_dependencies_error("resolve_genome_urls")

    def download_genome_references(*args, **kwargs):
        _missing_dependencies_error("download_genome_references")

    def check_command_availability(*args, **kwargs):
        _missing_dependencies_error("check_command_availability")

    def extract_region_tf_pairs(*args, **kwargs):
        _missing_dependencies_error("extract_region_tf_pairs")

    def build_gene_tss_dict(*args, **kwargs):
        _missing_dependencies_error("build_gene_tss_dict")

    def parse_region_name(*args, **kwargs):
        _missing_dependencies_error("parse_region_name")

    def build_pyranges_for_regions(*args, **kwargs):
        _missing_dependencies_error("build_pyranges_for_regions")

    def build_pyranges_for_genes(*args, **kwargs):
        _missing_dependencies_error("build_pyranges_for_genes")

    def build_region_gene_pairs(*args, **kwargs):
        _missing_dependencies_error("build_region_gene_pairs")

    def construct_region_tf_gene_triplets(*args, **kwargs):
        _missing_dependencies_error("construct_region_tf_gene_triplets")

    def rhg_to_rh_indexing(*args, **kwargs):
        _missing_dependencies_error("rhg_to_rh_indexing")

    def make_multiomic_dataset(*args, **kwargs):
        _missing_dependencies_error("make_multiomic_dataset")

    def pipeline_setup_multiview(*args, **kwargs):
        _missing_dependencies_error("pipeline_setup_multiview")

    def preprocessing_pipeline(*args, **kwargs):
        _missing_dependencies_error("preprocessing_pipeline")



from __future__ import annotations

from typing import Any, Dict

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from beartype import beartype
from numpyro.handlers import scale

from .RNA_utils import prepare_model_input, sample_prior
from .utils import solve_DE, sort_times_over_all_cells


@beartype
def mlp(
    params: Dict,
    x: Any,
) -> Any:
    """
    One or Multilayer Perceptron (MLP) with residual connections.

    Args:
        params: Dictionary containing neural network parameters (weights and biases).
        x: Input data array.

    Returns:
        Output array after passing through the MLP.
    """

    # First hidden layer with residual connections
    out = jnp.dot(x, params["W"]) + params["b"]
    out = jax.nn.softplus(out)

    return out


# Define the ODE function using the MLP
@beartype
def dstate_dt(t, state, args):
    """
    Compute the derivative of the state vector for the coupled system using a neural network.

    Args:
        t: Time scalar.
        state: State vector [u_1, ..., u_G, s_1, ..., s_G].
        args: Tuple containing parameters (G, beta_g, gamma_g, nn_params, T_ON).

    Returns:
        Derivative of the state vector
    """

    alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices, T_ON = args
    u = state[..., 0]  # Unspliced counts for all genes.
    s = state[..., 1]  # Spliced counts for all genes.

    # Clip state values to increase numerical stability:
    u = jnp.clip(u, 0, 1e3)
    s = jnp.clip(s, 0, 1e3)

    # Compute transcription rates using the neural network
    alpha = alpha_0 * jnp.clip(mlp(nn_params, s[:, tf_indices]), 0, 1e3)  # Shape: (G,)

    du_dt = alpha - beta_g * u
    ds_dt = beta_g * u - gamma_g * s

    dstate_dt = jnp.stack([du_dt, ds_dt], axis=-1)  # Shape: (2G,)
    return dstate_dt


# Define the complete NumPyro model
@beartype
def model(
    data: Any,
    M_c: Any,
    batch_index: Any,
    tf_indices: Any,
    total_num_cells: int,
    return_alpha: bool = False,
    n_batch: int = 1,
    Tmax_alpha: float = 50.0,
    Tmax_beta: float = 1.0,
    splicing_rate_alpha_hyp_prior_alpha: float = 20.0,
    splicing_rate_alpha_hyp_prior_mean: float = 5.0,
    splicing_rate_mean_hyp_prior_alpha: float = 10.0,
    splicing_rate_mean_hyp_prior_mean: float = 1.0,
    degradation_rate_alpha_hyp_prior_alpha: float = 20.0,
    degradation_rate_alpha_hyp_prior_mean: float = 5.0,
    degradation_rate_mean_hyp_prior_alpha: float = 10.0,
    degradation_rate_mean_hyp_prior_mean: float = 1.0,
    transcription_rate_alpha_hyp_prior_alpha: float = 20.0,
    transcription_rate_alpha_hyp_prior_mean: float = 2.0,
    transcription_rate_mean_hyp_prior_alpha: float = 10.0,
    transcription_rate_mean_hyp_prior_mean: float = 5.0,
    lambda_alpha: float = 1.0,
    lambda_mean: float = 20.0,
    kappa_alpha: float = 1.0,
    kappa_mean: float = 1.0,
    detection_mean_hyp_prior_alpha: float = 1.0,
    detection_mean_hyp_prior_beta: float = 1.0,
    detection_hyp_prior_alpha: float = 10.0,
    detection_i_prior_alpha: float = 100.0,
    detection_gi_prior_alpha: float = 200.0,
    gene_add_alpha_hyp_prior_alpha: float = 9.0,
    gene_add_alpha_hyp_prior_beta: float = 3.0,
    gene_add_mean_hyp_prior_alpha: float = 1.0,
    gene_add_mean_hyp_prior_beta: float = 100.0,
    stochastic_v_ag_hyp_prior_alpha: float = 9.0,
    stochastic_v_ag_hyp_prior_beta: float = 3.0,
):
    """
    NumPyro model for coupled transcription and splicing dynamics.

    Args:
        data: Observed data array of shape (num_cells, num_genes, num_modalities).
        M_c: Number of cells in each metacell.
        batch_index: Array indicating batch assignments for each cell.
        tf_indices: Indices of genes that are TFs.
        total_num_cells: Number of cells in the full dataset.
        n_batch: Number of batches.
        return_alpha: Whether to recompute the transcription rate (alpha), outside the ODE and save it as a deterministic site.
        Tmax_alpha: Alpha parameter for Tmax prior.
        Tmax_beta: Beta parameter for Tmax prior.
        ... (other hyperparameters for priors)
        key: Random number generator key.

    Returns:
        None. Defines the probabilistic model for inference.
    """

    num_cells = int(data.shape[0])
    num_genes = int(data.shape[1])
    num_modalities = int(data.shape[2])
    num_tfs = tf_indices.shape[0]
    batch_size = num_cells
    obs2sample = jax.nn.one_hot(
        batch_index, num_classes=n_batch
    )  # Shape: (num_cells, n_batch)

    # Splicing Rates for mRNA
    splicing_alpha = numpyro.sample(
        "splicing_alpha",
        dist.Gamma(
            splicing_rate_alpha_hyp_prior_alpha,
            splicing_rate_alpha_hyp_prior_alpha / splicing_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_alpha = jnp.clip(splicing_alpha, a_min=1e-2, a_max=1e3)

    splicing_mean = numpyro.sample(
        "splicing_mean",
        dist.Gamma(
            splicing_rate_mean_hyp_prior_alpha,
            splicing_rate_mean_hyp_prior_alpha / splicing_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    splicing_mean = jnp.clip(splicing_mean, a_min=1e-2, a_max=1e3)

    beta_g = numpyro.sample(
        "beta_g",
        dist.Gamma(splicing_alpha, splicing_alpha / splicing_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )

    # Degradation Rates for mRNA
    degradation_alpha = numpyro.sample(
        "degradation_alpha",
        dist.Gamma(
            degradation_rate_alpha_hyp_prior_alpha,
            degradation_rate_alpha_hyp_prior_alpha
            / degradation_rate_alpha_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_alpha = degradation_alpha + 0.001  # Prevent zero

    degradation_mean = numpyro.sample(
        "degradation_mean",
        dist.Gamma(
            degradation_rate_mean_hyp_prior_alpha,
            degradation_rate_mean_hyp_prior_alpha
            / degradation_rate_mean_hyp_prior_mean,
            validate_args=True,
        ),
    )
    degradation_mean = jnp.clip(degradation_mean, a_min=1e-2, a_max=1e3)

    gamma_g = numpyro.sample(
        "gamma_g",
        dist.Gamma(
            degradation_alpha, degradation_alpha / degradation_mean, validate_args=True
        )
        .expand([num_genes])
        .to_event(1),
    )

    # Translation rate for proteins
    lamda = numpyro.sample(
        "lambda",
        dist.Gamma(lambda_alpha, lambda_alpha / lambda_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )
    # Degradation rate for proteins
    kappa = numpyro.sample(
        "kappa",
        dist.Gamma(kappa_alpha, kappa_alpha / kappa_mean, validate_args=True)
        .expand([num_genes])
        .to_event(1),
    )

    # Time Parameters
    Tmax = numpyro.sample("Tmax", dist.Gamma(Tmax_alpha, Tmax_beta, validate_args=True))

    t_c_loc = numpyro.sample("t_c_loc", dist.Gamma(1.0, 1.0 / 0.5, validate_args=True))
    t_c_scale = numpyro.sample(
        "t_c_scale", dist.Gamma(1.0, 1.0 / 0.25, validate_args=True)
    )
    t_c_scale = jnp.clip(t_c_scale, a_min=1e-2)  # Prevent too small scales

    with numpyro.plate("cells", batch_size):
        t_c = numpyro.sample("t_c", dist.Normal(t_c_loc, t_c_scale))

    T_c = numpyro.deterministic("T_c", t_c * Tmax)

    # Time at which to start from initial condition:
    t_ON = numpyro.sample(
        "t_ON", dist.Normal(t_c_loc - t_c_scale, t_c_scale, validate_args=True)
    )

    T_ON = Tmax * t_ON

    # ============= Expression model =============== #

    # scale of alpha
    alpha_0 = numpyro.sample(
        "alpha_0",
        dist.Gamma(0.5, 0.5, validate_args=True).expand([num_genes]).to_event(1),
    )

    # Initial Conditions for ODE, sampling only for spliced and unspliced
    initial_state = numpyro.sample(
        "initial_state",
        dist.Gamma(1.0, 1.0, validate_args=True).expand([1, num_genes, 2]).to_event(1),
    )

    # Neural Net parameters
    in_dim = num_tfs
    out_dim = num_genes

    W = numpyro.sample("W", dist.Normal(0, 0.1), sample_shape=(in_dim, out_dim))

    b = numpyro.sample("b", dist.Normal(0, 0.1), sample_shape=(out_dim,))

    # Organize parameters into a dictionary
    nn_params = {
        "W": W,
        "b": b,
    }

    # Prepare Parameters for ODE Solver
    params = (alpha_0, beta_g, gamma_g, lamda, kappa, nn_params, tf_indices, T_ON)

    # Get Ordered Time Vector
    all_times, time_indices, _ = sort_times_over_all_cells(T_c)

    # Solve the Coupled ODE
    predictions = solve_DE(
        ts=all_times.squeeze(),
        params=params,
        initial_state=initial_state,
        time_step=0.05,
        model=dstate_dt,
        dense=False,
    )

    predictions = numpyro.deterministic("predictions", predictions.astype(jnp.float32))

    predictions_rearranged = numpyro.deterministic(
        "predictions_rearranged", predictions[time_indices.ravel(), :]
    )

    mu_expression = jnp.clip(
        predictions_rearranged[..., :2].squeeze(1), a_min=10 ** (-5), a_max=10 ** (5)
    )

    if return_alpha:
        final_s_clipped = jnp.clip(mu_expression[..., 1], 0, 1e3)
        alpha_cg = jnp.clip(
            alpha_0 * mlp(nn_params, final_s_clipped[:, tf_indices]), 0, 1e3
        )
        alpha_cg = numpyro.deterministic("alpha_cg", alpha_cg)

    # ============= Detection efficiency of spliced and unspliced counts =============== #

    with numpyro.plate("cells", batch_size):
        detection_y_c = numpyro.sample(
            "detection_y_c",
            dist.Gamma(detection_hyp_prior_alpha, detection_hyp_prior_alpha),
        )

    # ===================== Expected expression ======================= #

    mu = numpyro.deterministic("mu", mu_expression * detection_y_c[:, None, None] * M_c)

    # ===================== DATA likelihood ======================= #

    # Likelihood (sampling distribution)
    # We apply a scale factor so that the negative log-likelihood is multiplied
    # by (total_num_cells / batch_size). That ensures an unbiased gradient estimate.
    scale_factor = total_num_cells / batch_size
    with scale(scale=scale_factor):
        data_target = numpyro.sample("data_target", dist.Poisson(mu), obs=data)
